#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Modular Agent Mixin.

Provides shared functionality amongst modular agents.
"""
from abc import ABC
from typing import Dict, Any, List

from parlai.core.agents import Agent
from parlai.core.build_data import modelzoo_path
from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.loader import load_agent_module
import parlai.utils.logging as logging


class ModularAgentMixin(Agent, ABC):
    """
    Modular agent mixin.

    Provides shared functionality across modular agents (seeker, BB3)
    """

    def init_shared_model(self, opt: Opt, top_agent: Agent):
        """
        Initialize a shared version of a top-level agent.

        This just makes sure that each "agent" has the same params, but different
        history objects.

        :param opt_key:
            which sub agent to create with the shared model.
        """
        opt.update(opt['override'])
        if 'model' in opt['override']:
            model_class = load_agent_module(opt['override']['model'])
        else:
            model_class = type(top_agent)
        shared = top_agent.share()
        shared['opt'] = opt
        return model_class(opt, shared)

    def get_subagent_opt(
        self,
        datapath: str,
        filename: str,
        specific_override_args: Dict[str, Any],
        general_override_args: Dict[str, Any],
    ) -> Opt:
        """
        Given an agent opt, construct the new opt for the agent.

        :param filename:
            opt path
        :param specific_override_args:
            args for the specific agent
        :param general_override_args:
            args specified for all agents
        """
        if not filename.endswith('.opt'):
            filename += '.opt'
        opt = Opt.load(modelzoo_path(datapath, filename))
        opt['override'] = {}
        blocklist_general = ['model', 'model_file', 'init_model']
        general_override_args['skip_generation'] = False

        # Remove the prefix for the model for the specific override args.
        specific_override_args = {
            '_'.join(k.split('_')[1:]): v for k, v in specific_override_args.items()
        }

        override_args = {**general_override_args, **specific_override_args}

        for k, v in override_args.items():
            if k not in blocklist_general and k in opt:
                logging.warning(f'Overriding {k} to {v} (old val: {opt[k]})')
                opt['override'][k] = v
            elif k in specific_override_args:
                logging.warning(f'Key {k} not originally in opt, setting to {v}')
                opt['override'][k] = v

        return opt

    def batch_act_search_query_generation(
        self,
        observations: Any,
        search_query_generation_observations: List[Message],
        search_indices: List[int],
        search_query_agent: Agent,
        search_knowledge_agent: Agent,
        inject_query_string: str,
    ) -> List[Message]:
        """
        Search Query Generator batch act.

        :param observations:
            list of observations
        :param search_indices:
            list of batch indices for which search is required.

        :return batch_reply:
            return the batch reply from the search query agent
        """
        batch_reply = [Message({}) for _ in range(len(observations))]
        search_queries = []
        if search_indices:
            batch_replies_with_search = search_query_agent.batch_act(
                [
                    o
                    for i, o in enumerate(search_query_generation_observations)
                    if i in search_indices
                ]
            )
            for i, reply in zip(search_indices, batch_replies_with_search):
                batch_reply[i] = reply
            search_queries = [o.get('text', '') for o in batch_reply]
            if inject_query_string:
                for i in range(len(search_queries)):
                    if search_queries[i]:
                        new_query = ' '.join([search_queries[i], inject_query_string])
                        search_queries[i] = new_query
                        batch_reply[i].force_set('text', new_query)
            logging.debug(f"Search Queries: {search_queries}")
        try:
            search_knowledge_agent.model_api.set_search_queries(search_queries)
        except AttributeError:
            # Gold Documents, most likely
            pass

        return batch_reply



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Test world metrics + world metrics handlers against dummy conversations.
"""

import unittest

from parlai.core.tod.tod_core import (
    STANDARD_API_NAME_SLOT,
    STANDARD_REQUIRED_KEY,
    STANDARD_OPTIONAL_KEY,
    TodStructuredRound,
    TodStructuredEpisode,
    TodAgentType,
    TOD_AGENT_TYPE_TO_PREFIX,
)
from parlai.core.tod.world_metrics import TodMetrics
from parlai.core.tod.world_metrics_handlers import METRICS_HANDLER_CLASSES_TEST_REGISTRY

# Ignore lint on following line; want to have registered classes show up for tests
import projects.tod_simulator.world_metrics.extended_world_metrics  # noqa: F401

GOAL__SINGLE_ONE_KEY = [{STANDARD_API_NAME_SLOT: "name", "a": "1"}]
GOAL__SINGLE_THREE_KEYS = [
    {STANDARD_API_NAME_SLOT: "name", "a": "1", "b": "2", "c": "3"}
]
GOAL__HARD = [
    {
        STANDARD_API_NAME_SLOT: "otherName",
        "w": "1",
        "x": "2",
        "y": "3",
        "z": "will_be_missing",
        "diff": "right",
    }
]

API_CALL__NO_API_NAME_SLOT = {"random": "blah"}
API_CALL__API_NAME_DNE = {STANDARD_API_NAME_SLOT: "not_an_api_name"}
API_CALL__VALID_NAME_BUT_EMPTY = {STANDARD_API_NAME_SLOT: "name"}
API_CALL__SINGLE_ONE_KEY = GOAL__SINGLE_ONE_KEY[0]
API_CALL__SINGLE_ONE_KEY_WITH_OPT = {**GOAL__SINGLE_ONE_KEY[0], **{"c": "3"}}
API_CALL__SINGLE_ONE_KEY_WITH_OPT_AND_NONVALID = {
    **GOAL__SINGLE_ONE_KEY[0],
    **{"c": "3", "nonExistent": "blah"},
}
API_CALL__FUNKY_AGAINST_HARD = {
    STANDARD_API_NAME_SLOT: "otherName",
    "w": "1",
    "x": "2",
    "y": "3",
    "diff": "wrong",
}

API_SCHEMA__ONE_CALL_ONE_REQ_MATCH_ONE_KEY = [
    {
        STANDARD_API_NAME_SLOT: "name",
        STANDARD_REQUIRED_KEY: ["a"],
        STANDARD_OPTIONAL_KEY: [],
    }
]

API_SCHEMA__ONE_CALL_MATCH_THREE_KEYS = [
    {
        STANDARD_API_NAME_SLOT: "name",
        STANDARD_REQUIRED_KEY: ["a"],
        STANDARD_OPTIONAL_KEY: ["b", "c", "d"],
    }
]

API_SCHEMA__ONE_CALL_HARD = [
    {
        STANDARD_API_NAME_SLOT: "otherName",
        STANDARD_REQUIRED_KEY: ["w", "x"],
        STANDARD_OPTIONAL_KEY: ["y", "z", "diff"],
    }
]


class TodMetricsTestHelper:
    """
    Given a synthetic intermediate converesation, calculates the metrics for said
    conversation.
    """

    def __init__(self, e: TodStructuredEpisode):
        self.m = TodMetrics()
        self.m.handlers = [
            x() for x in METRICS_HANDLER_CLASSES_TEST_REGISTRY
        ]  # run on ALL
        self.e = e

    def _process(self, t: TodAgentType, text: str):
        self.m.handle_message({"text": f"{TOD_AGENT_TYPE_TO_PREFIX[t]}{text}"}, t)

    def run(self):
        self._process(TodAgentType.API_SCHEMA_GROUNDING_AGENT, self.e.api_schemas_utt)
        self._process(TodAgentType.GOAL_GROUNDING_AGENT, self.e.goal_calls_utt)

        for r in self.e.rounds:
            self._process(TodAgentType.USER_UTT_AGENT, r.user_utt)
            self._process(TodAgentType.API_CALL_AGENT, r.api_call_utt)
            self._process(TodAgentType.API_RESP_AGENT, r.api_resp_utt)
            self._process(TodAgentType.SYSTEM_UTT_AGENT, r.sys_utt)

        self.m.episode_reset()

    def report(self):
        return self.m.report()


class TestApiGoalHitMetricsHandler(unittest.TestCase):
    def __helper(self, api_schemas_machine, goal_calls_machine, single_turn_api_call):
        e = TodStructuredEpisode(
            api_schemas_machine=api_schemas_machine,
            goal_calls_machine=goal_calls_machine,
            rounds=[TodStructuredRound(api_call_machine=single_turn_api_call)],
        )
        helper = TodMetricsTestHelper(e)
        helper.run()
        result = helper.report()
        return result

    def test_one_goal_only_req(self):
        result = self.__helper(
            api_schemas_machine=API_SCHEMA__ONE_CALL_ONE_REQ_MATCH_ONE_KEY,
            goal_calls_machine=GOAL__SINGLE_ONE_KEY,
            single_turn_api_call=API_CALL__SINGLE_ONE_KEY,
        )
        self.assertAlmostEqual(result["all_goals_hit"], 1)
        self.assertAlmostEqual(result["all_goals_hit_turn_count"], 1)
        self.assertAlmostEqual(result["all_goals_fractional_hit"], 1)
        self.assertAlmostEqual(result["all_goals_slot_precision"], 1)
        self.assertAlmostEqual(result["all_goals_slot_recall"], 1)

        self.assertAlmostEqual(result["req_goals_hit"], 1)
        self.assertAlmostEqual(result["req_goals_hit_turn_count"], 1)
        self.assertAlmostEqual(result["req_goals_fractional_hit"], 1)
        self.assertAlmostEqual(result["req_goals_slot_precision"], 1)
        self.assertAlmostEqual(result["req_goals_slot_recall"], 1)

    def test_one_goal_api_name_missing_slots(self):
        result = self.__helper(
            api_schemas_machine=API_SCHEMA__ONE_CALL_ONE_REQ_MATCH_ONE_KEY,
            goal_calls_machine=GOAL__SINGLE_ONE_KEY,
            single_turn_api_call=API_CALL__VALID_NAME_BUT_EMPTY,
        )
        self.assertAlmostEqual(result["all_goals_hit"], 0)
        self.assertAlmostEqual(result["all_goals_hit_turn_count"], 0)
        self.assertAlmostEqual(result["all_goals_fractional_hit"], 0)
        self.assertAlmostEqual(result["all_goals_slot_precision"], 1)  # api_name
        self.assertAlmostEqual(result["all_goals_slot_recall"], 0.5)

        self.assertAlmostEqual(result["req_goals_hit"], 0)
        self.assertAlmostEqual(result["req_goals_hit_turn_count"], 0)
        self.assertAlmostEqual(result["req_goals_fractional_hit"], 0)
        self.assertAlmostEqual(result["req_goals_slot_precision"], 1)
        self.assertAlmostEqual(result["req_goals_slot_recall"], 0.5)

    def test_one_goal_with_opts(self):
        result = self.__helper(
            api_schemas_machine=API_SCHEMA__ONE_CALL_MATCH_THREE_KEYS,
            goal_calls_machine=GOAL__SINGLE_THREE_KEYS,
            single_turn_api_call=API_CALL__SINGLE_ONE_KEY,
        )
        self.assertAlmostEqual(result["all_goals_hit"], 0)
        self.assertAlmostEqual(result["all_goals_hit_turn_count"], 0)
        self.assertAlmostEqual(result["all_goals_fractional_hit"], 0)
        self.assertAlmostEqual(result["all_goals_slot_precision"], 1)
        self.assertAlmostEqual(result["all_goals_slot_recall"], 0.5)

        self.assertAlmostEqual(result["req_goals_hit"], 1)
        self.assertAlmostEqual(result["req_goals_hit_turn_count"], 1)
        self.assertAlmostEqual(result["req_goals_fractional_hit"], 1)
        self.assertAlmostEqual(result["req_goals_slot_precision"], 1)
        self.assertAlmostEqual(result["req_goals_slot_recall"], 1)

    def test_hard_case(self):
        result = self.__helper(
            api_schemas_machine=API_SCHEMA__ONE_CALL_HARD,
            goal_calls_machine=GOAL__HARD,
            single_turn_api_call=API_CALL__FUNKY_AGAINST_HARD,
        )
        self.assertAlmostEqual(result["all_goals_hit"], 0)
        self.assertAlmostEqual(result["all_goals_hit_turn_count"], 0)
        self.assertAlmostEqual(result["all_goals_fractional_hit"], 0)
        self.assertAlmostEqual(result["all_goals_slot_precision"], 0.8)
        self.assertAlmostEqual(result["all_goals_slot_recall"], 2.0 / 3.0)

        self.assertAlmostEqual(result["req_goals_hit"], 1)
        self.assertAlmostEqual(result["req_goals_hit_turn_count"], 1)
        self.assertAlmostEqual(result["req_goals_fractional_hit"], 1)
        self.assertAlmostEqual(result["req_goals_slot_precision"], 0.6)
        self.assertAlmostEqual(result["req_goals_slot_recall"], 1)


class TestApiCallMalformedMetricsHandler(unittest.TestCase):
    def __helper(self, single_turn_api_call):
        e = TodStructuredEpisode(
            api_schemas_machine=API_SCHEMA__ONE_CALL_MATCH_THREE_KEYS,
            rounds=[TodStructuredRound(api_call_machine=single_turn_api_call)],
        )
        helper = TodMetricsTestHelper(e)
        helper.run()
        return helper.report()

    def test_no_api_name_slot(self):
        result = self.__helper(API_CALL__NO_API_NAME_SLOT)
        self.assertEqual(result["apiCall_wellFormed"], 0)
        self.assertEqual(result["apiCall_hasSlotsButNoApiNameSlot_count"], 1)

    def test_api_name_DNE(self):
        result = self.__helper(API_CALL__API_NAME_DNE)
        self.assertEqual(result["apiCall_wellFormed"], 0)
        self.assertEqual(result["apiCall_methodDNE_count"], 1)

    def test_missing_required_slot(self):
        result = self.__helper(API_CALL__VALID_NAME_BUT_EMPTY)
        self.assertEqual(result["apiCall_wellFormed"], 0)
        self.assertEqual(result["apiCall_missingRequiredSlot_count"], 1)

    def test_has_single_required_slot(self):
        result = self.__helper(API_CALL__SINGLE_ONE_KEY)
        self.assertEqual(result["apiCall_wellFormed"], 1)
        self.assertEqual(result["apiCall_wellFormed_count"], 1)

    def test_has_valid_optional_slot(self):
        result = self.__helper(API_CALL__SINGLE_ONE_KEY_WITH_OPT)
        self.assertEqual(result["apiCall_wellFormed"], 1)
        self.assertEqual(result["apiCall_wellFormed_count"], 1)

    def test_has_invalid_extra_slots(self):
        result = self.__helper(API_CALL__SINGLE_ONE_KEY_WITH_OPT_AND_NONVALID)
        self.assertEqual(result["apiCall_wellFormed"], 0)
        self.assertEqual(result["apiCall_hasExtraParams_count"], 1)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import random
from typing import List
from parlai.core.message import Message
from parlai.core.mutators import register_mutator, ManyEpisodeMutator


@register_mutator("ltm_mutator")
class LongTermMemoryMutator(ManyEpisodeMutator):
    """
    Replaces the episode labels in messages with "personal_knwowledge".

    Episodes are flattened to ensure dialogue history is maintained appropriately.
    """

    def many_episode_mutation(self, episode: List[Message]) -> List[List[Message]]:
        history = []
        for message in episode:
            if message['text'] == '__SILENCE__':
                continue
            history.append(message.pop('text'))
            message['text'] = '\n'.join(history)
            labels = message.pop('labels')
            message['labels'] = ["personal_knowledge"]
            yield [message]
            history.append(random.choice(labels))



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
from parlai.core.build_data import DownloadableFile
import os
import jsonlines
from collections import defaultdict
from sklearn.model_selection import train_test_split

RANDOM_SEED = 123

RESOURCES = [
    DownloadableFile(
        'http://zissou.infosci.cornell.edu/convokit/datasets/friends-corpus/friends-corpus.zip',
        'friends-corpus.zip',
        '51ae80ce345212839d256b59b4982e9b40229ff6049115bd54d885a285d2b921',
        zipped=True,
    )
]


def generate_folds(dpath):
    """
    Generate Data Folds based on the scene id.
    """
    datafile = os.path.join(dpath, 'friends-corpus/utterances.jsonl')
    train_datafile = os.path.join(dpath, 'train.jsonl')
    valid_datafile = os.path.join(dpath, 'valid.jsonl')
    test_datafile = os.path.join(dpath, 'test.jsonl')

    # Load the dataset
    conversations = defaultdict(list)
    with jsonlines.open(datafile) as reader:
        for utterance in reader:
            text = utterance['text']
            speaker = utterance['speaker']
            conversation_id = utterance['conversation_id']

            if speaker != 'TRANSCRIPT_NOTE':
                conversations[conversation_id].append(
                    {
                        "text": text,
                        "speaker": speaker,
                        "conversation_id": conversation_id,
                    }
                )

    # Split the dataset into 80% train, 10% valid, 10% test
    train, valid_and_test = train_test_split(
        list(conversations.keys()), test_size=0.2, random_state=RANDOM_SEED
    )
    valid, test = train_test_split(
        valid_and_test, test_size=0.5, random_state=RANDOM_SEED
    )

    # Save the data folds into separate files
    with jsonlines.open(train_datafile, mode='w') as writer:
        for conversation_id in train:
            for utterance in conversations[conversation_id]:
                writer.write(utterance)

    with jsonlines.open(valid_datafile, mode='w') as writer:
        for conversation_id in valid:
            for utterance in conversations[conversation_id]:
                writer.write(utterance)

    with jsonlines.open(test_datafile, mode='w') as writer:
        for conversation_id in test:
            for utterance in conversations[conversation_id]:
                writer.write(utterance)


def build(opt):
    dpath = os.path.join(opt['datapath'], 'Friends')
    version = '1.00'

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        generate_folds(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Tests teachers + agent implementations, assuming parser to conversations format has
already been done and teachers/agents already created.

`test_agents.py` includes functions for generating the raw data used in this file as
well as the data parser.
"""

import unittest

import copy
import parlai.core.tod.tod_core as tod_core
import parlai.core.tod.tod_test_utils.test_agents as test_agents


class TestTodAgentsAndTeachersBase(unittest.TestCase):
    """
    Base class with convenience functions for setting up agents, dumping text, etc.
    """

    def setup_agent_or_teacher(self, class_type, round_opt, opt):
        full_opts = {**round_opt, **opt}
        full_opts["datatype"] = "DUMMY"
        full_opts["datafile"] = "DUMMY"
        full_opts["episodes_randomization_seed"] = -1  # no random here
        return class_type(full_opts)

    def dump_single_utt_per_episode_agent_text(self, class_type, round_opt, opt):
        """
        Continuously dumps data from an agent until it's done.
        """
        agent = self.setup_agent_or_teacher(class_type, round_opt, opt)
        result = []
        while not agent.epoch_done():
            result.append(agent.act()["text"])
            agent.reset()
        return result

    def dump_teacher_text(self, class_type, round_opt, opt):
        """
        Array where [episode_idx][turn_idx][text=0,label=1]
        """
        teacher = self.setup_agent_or_teacher(class_type, round_opt, opt)
        data = []
        here = []
        for x, new in teacher.setup_data("dummy"):
            if new and len(here) > 0:
                data.append(copy.deepcopy(here))
                here = []
            here.append([x["text"], x["label"]])
        if len(here) > 0:
            data.append(here)
        return data

    def _test_roundDataCorrect(self):
        """
        Convenience function that runs on different episode setups.

        Prefix with `_` since not all tests necessarily need this
        """
        self._test_roundDataCorrect_helper(test_agents.EPISODE_SETUP__UTTERANCES_ONLY)
        self._test_roundDataCorrect_helper(test_agents.EPISODE_SETUP__SINGLE_API_CALL)
        self._test_roundDataCorrect_helper(test_agents.EPISODE_SETUP__MULTI_ROUND)
        self._test_roundDataCorrect_helper(test_agents.EPISODE_SETUP__MULTI_EPISODE)

    def _test_roundDataCorrect_helper(self, config):
        """
        Implement this in downstream classes to define what is "correct" for a round (Ie
        checking serialization data for a given class vs only checking utterances)
        """
        raise RuntimeError("Not implemented")


class TestSystemTeacher(TestTodAgentsAndTeachersBase):
    def test_apiSchemas_with_yesApiSchemas(self):
        """
        Tests to make sure that data from first turn is correct when we include API
        Schemas.
        """
        values = self.dump_teacher_text(
            test_agents.SystemTeacher,
            test_agents.EPISODE_SETUP__SINGLE_API_CALL,
            {"api_schemas": True},
        )
        self.assertEqual(
            values[0][0][0],
            "APIS: "
            + tod_core.SerializationHelpers.list_of_maps_to_str(
                test_agents.make_api_schemas_machine(2)
            ),
        )

    def test_apiSchemas_with_noApiSchemas(self):
        """
        Tests to make sure that data from first turn is correct when we do not include
        API Schemas.
        """
        values = self.dump_teacher_text(
            test_agents.SystemTeacher,
            test_agents.EPISODE_SETUP__SINGLE_API_CALL,
            {"api_schemas": False},
        )
        self.assertEqual(values[0][0][0], "APIS: ")

    def _test_roundDataCorrect_helper(self, config):
        max_rounds = config[test_agents.TEST_NUM_ROUNDS_OPT_KEY]
        values = self.dump_teacher_text(test_agents.SystemTeacher, config, {})
        for episode_idx, episode in enumerate(values):
            utts = test_agents.get_round_utts(episode_idx, max_rounds)
            comp = []
            for utt in utts:
                comp.append([utt[0], utt[1]])
                comp.append([utt[2], utt[3]])
            # Skip grounding turn cause we check it in the other teachers
            self.assertEqual(episode[1:], comp)

    def test_roundDataCorrect(self):
        self._test_roundDataCorrect()


class TestUserTeacher(TestTodAgentsAndTeachersBase):
    def _test_roundDataCorrect_helper(self, config):
        """
        Make sure that all of the User teacher data is correct relative to ground truth,
        including grounding turn.
        """
        max_rounds = config[test_agents.TEST_NUM_ROUNDS_OPT_KEY]
        values = self.dump_teacher_text(test_agents.UserSimulatorTeacher, config, {})
        for episode_idx, episode in enumerate(values):
            utts = test_agents.get_round_utts(episode_idx, max_rounds)
            comp = []
            comp.append(
                [
                    "GOAL: "
                    + tod_core.SerializationHelpers.list_of_maps_to_str(
                        test_agents.make_goal_calls_machine(max_rounds)
                    ),
                    utts[0][0],
                ]
            )
            last_sys = utts[0][3]
            for i in range(1, len(utts)):
                comp.append([last_sys, utts[i][0]])
                last_sys = utts[i][3]
            self.assertEqual(episode, comp)

    def test_roundDataCorrect(self):
        self._test_roundDataCorrect()


class TestGoalAgent(TestTodAgentsAndTeachersBase):
    def _test_roundDataCorrect_helper(self, config):
        """
        Make sure goal agent data is correct with (possibly) multiple goals.
        """
        max_rounds = config[test_agents.TEST_NUM_ROUNDS_OPT_KEY]
        max_episodes = config[test_agents.TEST_NUM_EPISODES_OPT_KEY]
        values = self.dump_single_utt_per_episode_agent_text(
            test_agents.GoalAgent, config, {}
        )

        goal_text = [
            "GOAL: "
            + tod_core.SerializationHelpers.list_of_maps_to_str(
                test_agents.make_goal_calls_machine(max_rounds)
            )
            for _ in range(max_episodes)
        ]

        self.assertEqual(values, goal_text)

    def test_roundDataCorrect(self):
        self._test_roundDataCorrect()


class TestApiSchemaAgent(TestTodAgentsAndTeachersBase):
    def _test_roundDataCorrect_helper(self, config):
        """
        Make sure api schema information is correct with (possibly) multiple goals.
        """
        max_rounds = config[test_agents.TEST_NUM_ROUNDS_OPT_KEY]
        max_episodes = config[test_agents.TEST_NUM_EPISODES_OPT_KEY]
        values = self.dump_single_utt_per_episode_agent_text(
            test_agents.ApiSchemaAgent, config, {}
        )

        apis_texts = [
            "APIS: "
            + tod_core.SerializationHelpers.list_of_maps_to_str(
                test_agents.make_api_schemas_machine(max_rounds)
            )
            for _ in range(max_episodes)
        ]

        self.assertEqual(values, apis_texts)

    def test_roundDataCorrect(self):
        self._test_roundDataCorrect()


class TestSingleGoalAgent(TestTodAgentsAndTeachersBase):
    def _test_roundDataCorrect_helper(self, config):
        """
        Make sure single goal agent correctly splits conversations with multiple goals
        into single goals for the agent.
        """
        max_rounds = config[test_agents.TEST_NUM_ROUNDS_OPT_KEY]
        max_episodes = config[test_agents.TEST_NUM_EPISODES_OPT_KEY]
        values = self.dump_single_utt_per_episode_agent_text(
            test_agents.SingleGoalAgent, config, {}
        )

        goal_text = []
        for _ in range(max_episodes):
            goals = test_agents.make_goal_calls_machine(max_rounds)
            for x in goals:
                goal_text.append(
                    "GOAL: " + tod_core.SerializationHelpers.list_of_maps_to_str([x])
                )

        self.assertEqual(values, goal_text)

    def test_roundDataCorrect(self):
        self._test_roundDataCorrect()


class TestSingleApiSchemaAgent(TestTodAgentsAndTeachersBase):
    def _test_roundDataCorrect_helper(self, config):
        """
        Make sure single api schema agent correctly splits conversations with multiple
        goals into single goals for the agent.
        """
        max_rounds = config[test_agents.TEST_NUM_ROUNDS_OPT_KEY]
        max_episodes = config[test_agents.TEST_NUM_EPISODES_OPT_KEY]
        values = self.dump_single_utt_per_episode_agent_text(
            test_agents.SingleApiSchemaAgent, config, {}
        )

        apis_text = []
        for _ in range(max_episodes):
            apis = test_agents.make_api_schemas_machine(max_rounds)
            for x in apis:
                apis_text.append(
                    "APIS: " + tod_core.SerializationHelpers.list_of_maps_to_str([x])
                )
        self.assertEqual(values, apis_text)

    def test_roundDataCorrect(self):
        self._test_roundDataCorrect()


class TestSingleGoalWithSingleApiSchemaAgent(TestTodAgentsAndTeachersBase):
    """
    Make sure the SingleGoal + SingleApiSchema agents correspond.
    """

    def _test_roundDataCorrect_helper(self, config):
        goals = self.dump_single_utt_per_episode_agent_text(
            test_agents.SingleGoalAgent, config, {}
        )
        apis = self.dump_single_utt_per_episode_agent_text(
            test_agents.SingleApiSchemaAgent, config, {}
        )

        for i in range(len(goals)):
            goal = tod_core.SerializationHelpers.str_to_goals(goals[i][len("GOALS:") :])
            api = tod_core.SerializationHelpers.str_to_api_schemas(
                apis[i][len("APIS:") :]
            )
            self.assertEqual(
                goal[0].get("api_name", None), api[0].get("api_name", None)
            )

    def test_roundDataCorrect(self):
        self._test_roundDataCorrect()


class TestLowShot(TestTodAgentsAndTeachersBase):
    FEW_SHOT_SAMPLES = [0, 1, 5, 15]
    PERCENTAGES = [0, 0.1, 0.3, 0.5]

    def setup_agent_or_teacher(self, class_type, round_opt, opt):
        full_opts = {**round_opt, **opt}
        full_opts["datatype"] = "DUMMY"
        full_opts["datafile"] = "DUMMY"
        return class_type(full_opts)

    def test_few_shot_lengths_correct(self):
        def helper(n_shot):
            values = self.dump_teacher_text(
                test_agents.SystemTeacher,
                test_agents.EPISODE_SETUP__MULTI_EPISODE_BS,
                {"episodes_randomization_seed": 0, "n_shot": n_shot},
            )
            self.assertEqual(len(values), n_shot)

        for i in self.FEW_SHOT_SAMPLES:
            helper(i)

    def _test_subsets(self, data_dumps):
        for i in range(len(data_dumps) - 1):
            small = data_dumps[i]
            larger = data_dumps[i + 1]
            for i, episode in enumerate(small):
                self.assertEqual(episode, larger[i])

    def test_few_shot_subset(self):
        """
        Make sure specifying few-shot by n-shot works correctly.
        """

        def helper(n_shot, seed):
            return self.dump_teacher_text(
                test_agents.SystemTeacher,
                test_agents.EPISODE_SETUP__MULTI_EPISODE,
                {"episodes_randomization_seed": seed, "n_shot": n_shot},
            )

        data_dumps_seed_zero = [helper(i, 0) for i in self.FEW_SHOT_SAMPLES]
        self._test_subsets(data_dumps_seed_zero)
        data_dumps_seed_three = [helper(i, 3) for i in self.FEW_SHOT_SAMPLES]
        self._test_subsets(data_dumps_seed_three)
        self.assertNotEqual(data_dumps_seed_zero[-1], data_dumps_seed_three[-1])

    def test_percent_shot_lengths_correct(self):
        """
        Make sure specifying few-shot by percentages works correctly.
        """

        def helper(percent_shot, correct):
            values = self.dump_teacher_text(
                test_agents.SystemTeacher,
                test_agents.EPISODE_SETUP__MULTI_EPISODE_BS,  # 35 episodes
                {"episodes_randomization_seed": 0, "percent_shot": percent_shot},
            )
            self.assertEqual(len(values), correct)

        helper(0, 0)
        helper(0.1, 3)
        helper(0.3, 10)

    def test_percent_shot_subset(self):
        """
        Make sure specifying few-shot by percentages works correctly.
        """

        def helper(percent_shot, seed):
            return self.dump_teacher_text(
                test_agents.SystemTeacher,
                test_agents.EPISODE_SETUP__MULTI_EPISODE_BS,  # 35 episodes
                {"episodes_randomization_seed": seed, "percent_shot": percent_shot},
            )

        data_dumps_seed_zero = [helper(i, 0) for i in self.PERCENTAGES]
        self._test_subsets(data_dumps_seed_zero)
        data_dumps_seed_three = [helper(i, 3) for i in self.PERCENTAGES]
        self._test_subsets(data_dumps_seed_three)

    def test_correct_throw_when_both_shots_defined(self):
        self.assertRaises(
            RuntimeError,
            self.dump_teacher_text,
            test_agents.SystemTeacher,
            test_agents.EPISODE_SETUP__MULTI_EPISODE_BS,  # 35 episodes
            {"episodes_randomization_seed": 0, "percent_shot": 0.3, "n_shot": 3},
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Knowledge grounding task built from Language Model Data.
"""
from abc import ABC, abstractmethod
import os
from typing import Tuple, List, Optional, Any, Dict
import random

from parlai.core.message import Message
from parlai.core.metrics import (
    F1Metric,
    BleuMetric,
    RougeMetric,
    AverageMetric,
    ExactMatchMetric,
    normalize_answer,
)
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.teachers import ChunkTeacher, ChunkOutput
import parlai.tasks.wizard_of_internet.constants as CONST
from parlai.utils.conversations import Conversations
from parlai.utils.data import DatatypeHelper
import parlai.utils.logging as logging
from parlai.tasks.wizard_of_wikipedia.agents import TOKEN_KNOWLEDGE, TOKEN_END_KNOWLEDGE

from projects.seeker.utils import extract_entities, remove_possible_title_from_text


class CopiedSubstringMetric(AverageMetric):
    @staticmethod
    def compute(guess: str, answers: List[str]):
        if guess is None or answers is None:
            return None
        guess = normalize_answer(guess)
        for a in answers:
            if guess in normalize_answer(a):
                return CopiedSubstringMetric(1)
        return CopiedSubstringMetric(0)


NO_KNOWLEDGE_FIELD = 'no_knowledge_text'


def _path(opt: Opt, fname: str):
    root = opt['root_dir']
    return os.path.join(root, fname)


class AbstractLMChunkTeacher(ChunkTeacher, ABC):
    """
    LM Chunk Teacher, base.

    Provides functionality for subclassed teachers for extracting data from world log
    files, generated via `generate_lm_data.py`
    """

    def __init__(self, opt: Opt, shared=None) -> None:
        """
        Separate validation data from training data.
        """
        root = opt['root_dir']
        valid_indices = []
        for s in opt['validation_data_indices'].split(','):
            if s:
                valid_indices.append(int(s))
        if not valid_indices:
            logging.warning(
                "Warning: No validation indices provided, all data will be used for training"
            )
        self.data_files = sorted(
            [f for f in os.listdir(root) if f.endswith('.jsonl')],
            key=lambda x: int(x.replace('.jsonl', '').split('_')[-1]),
        )
        if DatatypeHelper.fold(opt['datatype']) in ['valid', 'test']:
            self.data_files = [
                d for i, d in enumerate(self.data_files) if i in valid_indices
            ]
        else:
            self.data_files = [
                d for i, d in enumerate(self.data_files) if i not in valid_indices
            ]
        opt['datafile'] = root
        super().__init__(opt, shared)  # type: ignore

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        ChunkTeacher.add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('LM Task Arguments')
        agent.add_argument(
            '--root-dir',
            type=str,
            default=None,
            help="Root directory with generated world logs.",
        )
        agent.add_argument(
            '--validation-data-indices',
            type=str,
            default='',
            help="comma separated list of validation data indices",
        )
        agent.add_argument(
            '--shared-knowledge-entity',
            type='bool',
            help='If True, only include examples with an entity in the knowledge that is also in the label',
            default=False,
        )
        agent.add_argument(
            '--min-knowledge-length',
            type=int,
            help='minimum length of the knowledge required to include the example. Default -1 means no min length',
            default=5,
        )
        agent.add_argument(
            '--min-knowledge-overlap',
            type=float,
            help='minimum overlap between knowledge and target sentence. Default 0 means no overlap required.',
            default=0,
        )
        agent.add_argument(
            '--skip-empty-context',
            type=bool,
            help='whether to exclude examples with empty context.',
            default=False,
        )
        agent.add_argument(
            '--exclude-retrieved-docs',
            type='bool',
            default=False,
            help='specify to not include retrieved docs in the episodes. hopefully reduces memory footprint.',
        )
        return parser

    def _get_ep_from_turns(
        self, xturns: List[Message], yturns: List[Message]
    ) -> List[Message]:
        """
        Return an episode given xturns and yturns.

        :param xturns:
            list of context turns
        :param yturns:
            list of label turns

        :return episodes:
            return built episodes
        """
        eps = []
        for xturn, yturn in zip(xturns, yturns):
            turn = {}
            # standard fields
            turn['text'] = (
                xturn.get('text')
                .strip()
                .replace(TOKEN_KNOWLEDGE, f'{TOKEN_KNOWLEDGE} ')
                .replace(TOKEN_END_KNOWLEDGE, f' {TOKEN_END_KNOWLEDGE}')
            )
            # subtract 1 to get rid of newline char
            if len(turn['text'].split('\n')) == 1:
                # no actual context here...
                turn[NO_KNOWLEDGE_FIELD] = ''
            else:
                newline_correction = int(turn['text'].endswith('\n'))
                turn[NO_KNOWLEDGE_FIELD] = turn['text'][
                    : turn['text'].index(TOKEN_KNOWLEDGE) - newline_correction
                ]
            turn['labels'] = [yturn.get('text').strip()]
            turn['episode_done'] = False
            # my fields
            if not self.opt['exclude_retrieved_docs']:
                turn[CONST.RETRIEVED_DOCS] = yturn['retrieved_docs']
                turn[CONST.RETRIEVED_DOCS_TITLES] = [
                    d.split(' / ')[0] for d in yturn['retrieved_docs']
                ]
                turn[CONST.RETRIEVED_DOCS_URLS] = ['' for _ in yturn['retrieved_docs']]
                turn[CONST.SELECTED_DOCS] = [yturn['gold_doc']]
                turn[CONST.SELECTED_DOCS_TITLES] = [yturn['gold_doc'].split(' / ')[0]]
                turn[CONST.SELECTED_DOCS_URLS] = []
                turn[CONST.SELECTED_SENTENCES] = [yturn['knowledge']]
            else:
                turn[CONST.SELECTED_DOCS_TITLES] = [yturn['gold_doc'].split(' / ')[0]]
                turn[CONST.SELECTED_SENTENCES] = [yturn['knowledge']]

            turn['search_query'] = yturn['search_query']
            turn['f1_overlap'] = yturn['f1_overlap']
            if self.passes_filters(xturn, yturn):
                eps.append(self.finalize_message(turn))
        if eps:
            eps[-1].force_set('episode_done', True)
        return eps

    def passes_filters(self, xturn: Dict[str, Any], yturn: Dict[str, Any]) -> bool:
        """
        Subject example to various filters.

        Return whether the example passes all filters.

        :param xturn:
            context turn
        :param yturn:
            target/knowledge turn

        :return passes_filters:
            return whether the example passes the filters.
        """
        passes = True
        # Example filters
        knowledge = (
            yturn['knowledge']
            .replace(TOKEN_KNOWLEDGE, '')
            .replace(TOKEN_END_KNOWLEDGE, '')
            .strip()
        )
        if passes and self.opt['skip_empty_context']:
            doc_context_sentences = [
                s for s in xturn['text'].split(TOKEN_KNOWLEDGE)[0].split('\n') if s
            ]
            passes &= (
                len(doc_context_sentences)
            ) > 1  # All docs have <doc> token as their first line
        if passes and self.opt['min_knowledge_length'] > 0:
            passes &= len(knowledge.split(' ')) >= self.opt['min_knowledge_length']
        if passes and self.opt['min_knowledge_overlap'] > 0:
            assert 0 < self.opt['min_knowledge_overlap'] <= 1
            f1 = F1Metric.compute(yturn['text'].strip(), [knowledge])
            passes &= f1.value() >= self.opt['min_knowledge_overlap']
        if passes and self.opt['shared_knowledge_entity']:
            knol_ent = extract_entities(knowledge)
            if len(knol_ent) == 0:
                passes &= False
            label_ent = extract_entities(yturn.get('text'))
            ents = set(knol_ent).intersection(label_ent)
            if len(ents) == 0:
                passes &= False
        return passes

    @abstractmethod
    def finalize_message(self, msg: Dict[str, Any]) -> Message:
        """
        Finalize message.

        Override in subclasses to set text/labels appropriately.
        """

    def get_num_samples(self, opt: Opt) -> Tuple[int, int]:
        """
        Return the number of samples.

        Returns a tuple of (num_examples, num_episodes) based on the data split.
        """
        return len(self.data_files) * 1000, len(self.data_files) * 1000

    def get_fold_chunks(self, opt: Opt) -> List[int]:
        """
        Return a list of chunk IDs (integer).

        Given the datatype (train/test/valid), return the list of chunk IDs that
        correspond to that split.
        """
        return list(range(len(self.data_files)))

    def load_from_chunk(self, chunk_idx: int) -> List[ChunkOutput]:
        """
        Given the chunk index, load examples from that chunk.

        Return a list of tuples. The function `_create_message` will take these tuples
        to form the Message object that is returned by the teacher.
        """
        convs = Conversations(_path(self.opt, self.data_files[chunk_idx]))
        chunk = []
        for conv in convs:
            turns = [t for t in conv.turns if t.get('id') != 'context']
            ep = self._get_ep_from_turns(turns[::2], turns[1::2])
            if not ep:
                continue
            chunk += ep
        return chunk

    def create_message(self, queue_output: ChunkOutput, entry_idx=0) -> Message:
        """
        Given the tuple output of the queue, return an act.

        May depend on entry index if queue output is a multi-turn episode.
        """
        return queue_output


class ResponseTeacher(AbstractLMChunkTeacher):
    """
    Response teacher.

    For training the dialogue response component of SeeKeR.
    """

    def finalize_message(self, msg: Dict[str, Any]) -> Message:
        """
        Return message as is.
        """
        return Message(msg)


class ResponseNoKnowledgeTeacher(AbstractLMChunkTeacher):
    """
    A variant that removes the knowledge from the raw data.
    """

    def finalize_message(self, msg: Dict[str, Any]) -> Message:
        """
        Remove the knowledge component of the message.
        """
        msg['text'] = msg[NO_KNOWLEDGE_FIELD]
        return Message(msg)


class KnowledgeTeacher(AbstractLMChunkTeacher):
    """
    Knowledge Teacher.

    Input is Context. Target is selected __knowledge__.
    """

    def custom_evaluation(
        self,
        teacher_action: Message,
        labels: Optional[Tuple[str]],
        model_response: Message,
    ) -> None:
        """
        Various F1 metrics for the generated model response.
        """
        if not model_response.get('text'):
            # No response generated by model.
            return

        resp = model_response['text']
        # F1 metric over the *selected* knowledge.
        self.metrics.add(
            'knowledge_f1_docs',
            F1Metric.compute(resp, teacher_action[CONST.SELECTED_DOCS]),
        )
        self.metrics.add(
            'knowledge_f1_sentences',
            F1Metric.compute(resp, teacher_action[CONST.SELECTED_SENTENCES]),
        )

        # F1 Metrics over the *retrieved* docs.
        self.metrics.add(
            'f1_retrieved_docs',
            F1Metric.compute(resp, ' '.join(teacher_action[CONST.RETRIEVED_DOCS])),
        )
        self.metrics.add(
            'max_f1_retrieved_docs',
            F1Metric.compute(resp, teacher_action[CONST.RETRIEVED_DOCS]),
        )

        selected_doc_senetences = teacher_action[CONST.SELECTED_DOCS][0].split('\n')
        all_doc_senetences = []
        for doc in teacher_action[CONST.RETRIEVED_DOCS]:
            all_doc_senetences.extend(doc.split('\n'))

        self.metrics.add(
            'exact_copied_sentences', ExactMatchMetric.compute(resp, all_doc_senetences)
        )
        self.metrics.add(
            'max_substring_copied_sentences',
            CopiedSubstringMetric.compute(resp, all_doc_senetences),
        )
        self.metrics.add(
            'max_substring_copied_docs',
            CopiedSubstringMetric.compute(resp, teacher_action[CONST.RETRIEVED_DOCS]),
        )
        self.metrics.add(
            'substring_copied_docs',
            CopiedSubstringMetric.compute(
                resp, [''.join(teacher_action[CONST.RETRIEVED_DOCS])]
            ),
        )
        self.metrics.add(
            'max_f1_selected_docs_senetences',
            F1Metric.compute(resp, selected_doc_senetences),
        )
        self.metrics.add(
            'max_f1_docs_senetences', F1Metric.compute(resp, all_doc_senetences)
        )

        # N-gram matching metrics
        for k in range(1, 5):  # 1..4
            self.metrics.add(
                f'max_bleu_selected_docs_senetences-{k}',
                BleuMetric.compute(resp, selected_doc_senetences, k),
            )

        r1, r2, rL = RougeMetric.compute_many(resp, selected_doc_senetences)
        self.metrics.add('max_rouge_selected_docs_senetences_1', r1)
        self.metrics.add('max_rouge_selected_docs_senetences_2', r2)
        self.metrics.add('max_rouge_selected_docs_senetences_L', rL)

    def finalize_message(self, msg: Dict[str, Any]) -> Message:
        """
        Remove knowledge component of message, and move selected sentence to target.
        """
        msg['text'] = msg[NO_KNOWLEDGE_FIELD]
        msg['labels'] = [
            ' '.join(msg[CONST.SELECTED_SENTENCES])
            .replace(TOKEN_KNOWLEDGE, '')
            .replace(TOKEN_END_KNOWLEDGE, '')
            .strip()
        ]
        return Message(msg)


class SearchQueryTeacher(KnowledgeTeacher):
    """
    Search Query teacher.

    Given an input context, predict the title of the document from which the "gold"
    knowledge is derived.
    """

    @classmethod
    def add_cmdline_args(cls, parser, partial_opt=None):
        KnowledgeTeacher.add_cmdline_args(parser, partial_opt=partial_opt)
        agent = parser.add_argument_group("Selected Doc Titles Arguments")
        agent.add_argument(
            "--remove-possible-title",
            type=bool,
            default=True,
            help="Remove the length of characters from the beginning of the text that exactly match the title.",
        )
        agent.add_argument(
            "--random-trim-min-length",
            type=int,
            default=-1,
            help=(
                "If text is longer than this length (in chars), randomly trim from the beginning "
                "such that the text is at least this long (default -1, no trimming)."
            ),
        )
        return parser

    def __init__(self, opt, shared=None):
        self._rem_possible_title = opt['remove_possible_title']
        self._random_trim_min_len = opt['random_trim_min_length']
        super().__init__(opt, shared=shared)

    def _clean_text(self, txt, label):
        PRFX = '<doc>'
        if txt.startswith(PRFX):
            # removing the extra white-space character after the PRFX too.
            txt = txt[len(PRFX) + 1 :]

        if self._rem_possible_title:
            txt = remove_possible_title_from_text(txt, label)

        if self._random_trim_min_len > 0 and len(txt) > self._random_trim_min_len:
            trim_point_idx = random.randint(0, len(txt) - self._random_trim_min_len)
            txt = txt[trim_point_idx:]

        return txt

    def finalize_message(self, msg: Dict[str, Any]) -> Message:
        assert (
            len(msg[CONST.SELECTED_DOCS_TITLES]) == 1
        ), 'Only 1 doc must be selected by this teacher.'

        label = msg[CONST.SELECTED_DOCS_TITLES][0]
        msg['text'] = self._clean_text(msg[NO_KNOWLEDGE_FIELD], label)
        msg['labels'] = [label]
        return Message(msg)


class DefaultTeacher(ResponseTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Training script for ParlAI.

The standard way to train a model. After training, also computes
validation and test error.

The user must provide a model (with `--model`) and a task (with
`--task`).

## Examples

```shell
parlai train_model --model ir_baseline --task dialog_babi:Task:1 --model-file /tmp/model
parlai train_model --model seq2seq --task babi:Task10k:1 --model-file '/tmp/model' --batchsize 32 --learningrate 0.5
```
"""  # noqa: E501

# TODO List:
# * More logging (e.g. to files), make things prettier.
import copy
import random
import torch
import json
import os
import signal
from typing import Tuple

import numpy as np

import parlai.utils.logging as logging
from parlai.core.agents import create_agent, create_agent_from_shared
from parlai.core.exceptions import StopTrainException
from parlai.core.logs import TensorboardLogger, WandbLogger, ClearMLLogger
from parlai.core.metrics import Metric
from parlai.core.metrics import (
    aggregate_named_reports,
    aggregate_unnamed_reports,
    dict_report,
)
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser, print_announcements
from parlai.core.script import ParlaiScript, register_script
from parlai.core.worlds import create_task, World
from parlai.scripts.build_dict import build_dict, setup_args as setup_dict_args
from parlai.scripts.eval_model import get_task_world_logs
from parlai.utils.distributed import (
    sync_object,
    is_primary_worker,
    all_gather_list,
    is_distributed,
    get_rank,
    num_workers,
)
from parlai.utils.io import PathManager
from parlai.utils.misc import Timer, nice_report
from parlai.utils.world_logging import WorldLogger


def _num_else_inf(opt: Opt, key: str, distributed_warn=False):
    if opt[key] > 0:
        if distributed_warn and is_distributed():
            nicekey = '--' + key.replace('_', '-')
            logging.warning(
                f'Using {nicekey} in distributed mode can lead to slowdowns. '
                'See https://github.com/facebookresearch/ParlAI/pull/3379 for more info.'
            )
        value = opt[key]
    else:
        value = float('inf')
    return value


def setup_args(parser=None) -> ParlaiParser:
    """
    Build the ParlAI parser, adding command line args if necessary.

    :param ParlaiParser parser:
        Preexisting parser to append options to. Will be created if needed.

    :returns:
        the ParlaiParser with CLI options added.
    """
    if parser is None:
        parser = ParlaiParser(True, True, 'Train a model')
    train = parser.add_argument_group('Training Loop Arguments')
    train.add_argument(
        '-et',
        '--evaltask',
        help='task to use for valid/test (defaults to the one used for training)',
    )
    train.add_argument(
        '--final-extra-opt',
        type=str,
        default='',
        help="A '.opt' file that is used for final eval. Useful for setting skip-generation to false. 'datatype' must be included as part of the opt.",
    )
    train.add_argument(
        '--eval-batchsize',
        type=int,
        hidden=True,
        help='Eval time batch size (defaults to same as -bs)',
    )
    train.add_argument(
        '--eval-dynamic-batching',  # FIXME: see https://github.com/facebookresearch/ParlAI/issues/3367
        default=None,
        type='nonestr',
        choices={None, 'off', 'full', 'batchsort'},
        help=(
            'Set dynamic batching at evaluation time. Set to off for '
            'train-only dynamic batching. Set to none (default) to use same '
            'setting as --dynamic-batching.'
        ),
    )
    train.add_argument(
        '--num-workers',
        default=0,
        type=int,
        help='Number of background workers (training only)',
    )
    train.add_argument('--display-examples', type='bool', default=False, hidden=True)
    train.add_argument('-eps', '--num-epochs', type=float, default=-1)
    train.add_argument('-ttim', '--max-train-time', type=float, default=-1)
    train.add_argument(
        '-tstep',
        '--max-train-steps',
        '--max-lr-steps',
        type=int,
        default=-1,
        help='End training after n model updates',
    )
    train.add_argument('-ltim', '--log-every-n-secs', type=float, default=-1)
    train.add_argument(
        '-lstep',
        '--log-every-n-steps',
        type=int,
        default=50,
        help='Log every n training steps',
    )
    train.add_argument(
        '-vtim',
        '--validation-every-n-secs',
        type=float,
        default=-1,
        help='Validate every n seconds. Saves model to model_file '
        '(if set) whenever best val metric is found',
    )
    train.add_argument(
        '-vstep',
        '--validation-every-n-steps',
        type=int,
        default=-1,
        help='Validate every n training steps. Saves model to model_file '
        '(if set) whenever best val metric is found',
    )
    train.add_argument(
        '-stim',
        '--save-every-n-secs',
        type=float,
        default=-1,
        help='Saves the model to model_file.checkpoint after '
        'every n seconds (default -1, never).',
    )
    train.add_argument(
        '-sval',
        '--save-after-valid',
        type='bool',
        default=False,
        help='Saves the model to model_file.checkpoint after '
        'every validation (default %(default)s).',
    )
    train.add_argument(
        '-veps',
        '--validation-every-n-epochs',
        type=float,
        default=-1,
        help='Validate every n epochs. Saves model to model_file '
        '(if set) whenever best val metric is found',
    )
    train.add_argument(
        '-vme',
        '--validation-max-exs',
        type=int,
        default=-1,
        hidden=True,
        help='max examples to use during validation (default -1 uses all)',
    )
    train.add_argument(
        '--short-final-eval',
        default=False,
        hidden=True,
        type='bool',
        help='If true, obeys --validation-max-exs in the final '
        'validation and test evaluations.',
    )
    train.add_argument(
        '-vp',
        '--validation-patience',
        type=int,
        default=10,
        help=(
            'number of iterations of validation where result'
            ' does not improve before we stop training'
        ),
    )
    train.add_argument(
        '-vmt',
        '--validation-metric',
        default='accuracy',
        help='key into report table for selecting best validation',
    )
    train.add_argument(
        '-vmm',
        '--validation-metric-mode',
        type=str,
        choices=['max', 'min'],
        help='the direction in which to optimize the validation metric, i.e. maximize or minimize',
    )
    train.add_argument(
        '-vcut',
        '--validation-cutoff',
        type=float,
        default=1.0,
        hidden=True,
        help='value at which training will stop if exceeded by metric',
    )
    train.add_argument(
        '-lfc',
        '--load-from-checkpoint',
        type='bool',
        default=True,
        hidden=True,
        help='load model from checkpoint if available',
    )
    train.add_argument(
        '-vshare',
        '--validation-share-agent',
        default=False,
        hidden=True,
        help='use a shared copy of the agent for validation. '
        'this will eventually default to True, but '
        'currently defaults to False.',
    )
    train.add_argument(
        '-mcs',
        '--metrics',
        type=str,
        default='default',
        help='list of metrics to show/compute, e.g. all, default,'
        'or give a list split by , like '
        'ppl,f1,accuracy,hits@1,rouge,bleu'
        'the rouge metrics will be computed as rouge-1, rouge-2 and rouge-l',
    )
    train.add_argument(
        '-micro',
        '--aggregate-micro',
        type='bool',
        default=False,
        help='Report micro-averaged metrics instead of macro averaged metrics.',
        recommended=False,
    )
    train.add_argument(
        '--world-logs',
        type=str,
        default='',
        help='Saves a jsonl file of the world logs.'
        'Set to the empty string to not save at all.',
    )
    train.add_argument(
        '--save-format',
        type=str,
        default='conversations',
        choices=['conversations', 'parlai'],
    )
    train.add_argument('--seed', type=int, default=None)
    WorldLogger.add_cmdline_args(parser, partial_opt=None)
    TensorboardLogger.add_cmdline_args(parser, partial_opt=None)
    WandbLogger.add_cmdline_args(parser, partial_opt=None)
    ClearMLLogger.add_cmdline_args(parser, partial_opt=None)
    parser = setup_dict_args(parser)
    return parser


def set_seed(seed):
    random.seed(seed)
    torch.manual_seed(seed)


def load_eval_worlds(agent, opt, datatype):
    """
    Create a new eval world for the agent and the given opt.

    Overrides the datatype options for doing this.  Handles some magic
    overrides of other special options for the training script.

    :param Agent agent:
        The model being trained.

    :param Opt opt:
        The global CLI opts.

    :param string datatype:
        The new datatype.
    """

    if 'stream' in opt['datatype']:
        datatype += ':stream'
    opt = opt.copy()
    opt['datatype'] = datatype
    if opt.get('evaltask'):
        # if a different eval task is specified, use it.
        opt['task'] = opt['evaltask']
    if opt.get('eval_batchsize'):
        # override eval time batchsize
        opt['batchsize'] = opt['eval_batchsize']
    if opt.get('eval_dynamic_batching'):
        # FIXME: see issue tracked in https://github.com/facebookresearch/ParlAI/issues/3367
        # override eval time dynamic batching settings
        eval_dyn_batch = (
            None
            if opt['eval_dynamic_batching'] == 'off'
            else opt['eval_dynamic_batching']
        )
        opt['dynamic_batching'] = eval_dyn_batch

    tasks = opt['task'].split(',')
    worlds = []
    # possibly load agent
    if opt.get('validation_share_agent', False):
        valid_agent = create_agent_from_shared(agent.share())
    else:
        valid_agent = agent
    # create worlds
    for task in tasks:
        task_opt = opt.copy()  # copy opt since we edit the task
        task_opt['task'] = task
        valid_world = create_task(task_opt, valid_agent)
        worlds.append(valid_world)

    return worlds


class TrainLoop:
    """
    TrainLoop contains the core training loop logic.
    """

    def __init__(self, opt):
        # if python is called from a non-interactive shell, like a bash script,
        # it will by-default ignore SIGINTs, and KeyboardInterrupt exceptions are
        # not produced. This line brings them back
        signal.signal(signal.SIGINT, signal.default_int_handler)
        # Possibly load from checkpoint
        trainstats_suffix = '.trainstats'  # we might load training statistics from here
        if (
            opt['load_from_checkpoint']
            and opt.get('model_file')
            and PathManager.exists(opt['model_file'] + '.checkpoint')
        ):
            opt['init_model'] = opt['model_file'] + '.checkpoint'
            trainstats_suffix = '.checkpoint.trainstats'
        # Possibly build a dictionary (not all models do this).
        if not (opt.get('dict_file') or opt.get('model_file')):
            raise RuntimeError(
                'WARNING: For train_model, please specify either a '
                'model_file or dict_file.'
            )
        if 'dict_file' in opt:
            if opt['dict_file'] is None and opt.get('model_file'):
                opt['dict_file'] = opt['model_file'] + '.dict'
            logging.info("building dictionary first...")
            build_dict(opt, skip_if_built=True)

        # Create model and assign it to the specified task
        self.agent = create_agent(opt)
        self.agent.opt.log()
        self.world = create_task(opt, self.agent)
        # set up timers
        self.train_time = Timer()
        self.validate_time = Timer()
        self.log_time = Timer()
        self.save_time = Timer()

        self.parleys = 0
        self._train_steps = 0
        self._last_log_steps = 0
        self.update_freq = opt.get('update_freq', 1)

        self.max_num_epochs = _num_else_inf(opt, 'num_epochs', distributed_warn=True)
        self.max_train_time = _num_else_inf(
            opt, 'max_train_time', distributed_warn=True
        )
        self.max_train_steps = _num_else_inf(opt, 'max_train_steps')
        self.log_every_n_secs = _num_else_inf(
            opt, 'log_every_n_secs', distributed_warn=True
        )
        self.log_every_n_steps = _num_else_inf(opt, 'log_every_n_steps')
        self.val_every_n_secs = _num_else_inf(
            opt, 'validation_every_n_secs', distributed_warn=True
        )
        self.val_every_n_epochs = _num_else_inf(
            opt, 'validation_every_n_epochs', distributed_warn=True
        )
        self.val_every_n_steps = _num_else_inf(opt, 'validation_every_n_steps')
        self.save_every_n_secs = _num_else_inf(
            opt, 'save_every_n_secs', distributed_warn=True
        )

        # smart defaults for --validation-metric-mode
        if opt['validation_metric'] in {'loss', 'ppl', 'mean_rank'}:
            opt['validation_metric_mode'] = 'min'
        elif opt['validation_metric'] in {'accuracy', 'hits@1', 'hits@5', 'f1', 'bleu'}:
            opt['validation_metric_mode'] = 'max'
        if opt.get('validation_metric_mode') is None:
            opt['validation_metric_mode'] = 'max'

        self.last_valid_epoch = 0
        self._last_valid_steps = 0
        self.valid_optim = 1 if opt['validation_metric_mode'] == 'max' else -1
        self.train_reports = []
        self.valid_reports = []
        self.final_valid_report = {}
        self.final_test_report = {}
        self.final_extra_valid_report = {}
        self.best_valid = None

        self.impatience = 0
        self.saved = False
        self.valid_worlds = None
        self.opt = opt

        # we may have been preempted, make sure we note that amount
        self._preempted_epochs = 0.0
        if opt.get('model_file') and PathManager.exists(
            opt['model_file'] + trainstats_suffix
        ):
            # looks like we were preempted. make sure we load up our total
            # training stats, etc
            with PathManager.open(opt['model_file'] + trainstats_suffix) as ts:
                obj = json.load(ts)
                self.parleys = obj.get('parleys', 0)
                self._preempted_epochs = obj.get('total_epochs', 0)
                self.train_time.total = obj.get('train_time', 0)
                self._train_steps = obj.get('train_steps', 0)
                self.impatience = obj.get('impatience', 0)
                self.valid_reports = obj.get('valid_reports', [])
                if self.valid_reports:
                    self.last_valid_epoch = self.valid_reports[-1].get(
                        'total_epochs', 0.0
                    )
                self.train_reports = obj.get('train_reports', [])
                if 'best_valid' in obj:
                    self.best_valid = obj['best_valid']
                else:
                    # old method
                    if opt.get('model_file') and PathManager.exists(
                        opt['model_file'] + '.best_valid'
                    ):
                        with PathManager.open(
                            opt['model_file'] + ".best_valid", 'r'
                        ) as f:
                            x = f.readline()
                            self.best_valid = float(x)
                            f.close()

        if opt['tensorboard_log'] and is_primary_worker():
            self.tb_logger = TensorboardLogger(opt)
        if opt['wandb_log'] and is_primary_worker():
            model = self.agent.model if hasattr(self.agent, 'model') else None
            self.wb_logger = WandbLogger(opt, model)
        if opt['clearml_log'] and is_primary_worker():
            self.clearml_logger = ClearMLLogger(opt)

    def save_model(self, suffix=None):
        """
        Save the model to disk, possibly with a suffix.
        """
        if not self.opt.get('model_file'):
            # nothing to save to, just exit
            return

        fn = self.opt['model_file']
        if suffix:
            fn += suffix

        if not is_primary_worker():
            # never do IO as a non-primary worker
            if hasattr(self.agent, 'save_nonprimary'):
                self.agent.save_nonprimary(fn)
            return

        while True:
            # don't ever let a ctrl-c interrupt saving
            try:
                self.agent.save(fn)
                self._save_train_stats(suffix)
                if self.opt['wandb_log'] and self.opt["wandb_log_model"]:
                    self.wb_logger.log_model(fn)
                break
            except KeyboardInterrupt:
                pass

    def _save_train_stats(self, suffix=None):
        if not is_primary_worker():
            # never do IO as a non-primary worker
            return
        fn = self.opt.get('model_file', None)
        if not fn:
            return
        if suffix:
            fn += suffix
        fn += '.trainstats'
        with PathManager.open(fn, 'w') as f:
            json.dump(
                {
                    'parleys': self.parleys,
                    'train_time': self.train_time.time(),
                    'train_steps': self._train_steps,
                    'total_epochs': self._total_epochs,
                    'train_reports': self.train_reports,
                    'valid_reports': self.valid_reports,
                    'best_valid': self.best_valid,
                    'impatience': self.impatience,
                    'final_valid_report': dict_report(self.final_valid_report),
                    'final_test_report': dict_report(self.final_test_report),
                    'final_extra_valid_report': dict_report(
                        self.final_extra_valid_report
                    ),
                },
                f,
                indent=4,
            )

    def validate(self):
        """
        Perform a validation run, checking whether we should stop training.

        :return: boolean indicating whether training should stop
        :rtype: bool
        """
        opt = self.opt

        if self.valid_worlds is None:
            # we need to load the world now
            self.valid_worlds = load_eval_worlds(self.agent, opt, 'valid')

        # run evaluation on valid set
        valid_report = self._run_eval(
            self.valid_worlds, opt, 'valid', opt['validation_max_exs']
        )
        v = dict_report(valid_report)
        v['train_time'] = self.train_time.time()
        v['parleys'] = self.parleys
        v['train_steps'] = self._train_steps
        v['total_exs'] = self._total_exs
        v['total_epochs'] = self._total_epochs
        self.valid_reports.append(v)
        # logging
        if opt['tensorboard_log'] and is_primary_worker():
            valid_report['total_exs'] = self._total_exs
            self.tb_logger.log_metrics('valid', self.parleys, valid_report)
            # flush on a validation
            self.tb_logger.flush()
        if opt['wandb_log'] and is_primary_worker():
            valid_report['total_exs'] = self._total_exs
            self.wb_logger.log_metrics('valid', self.parleys, valid_report)

        if opt['clearml_log'] and is_primary_worker():
            valid_report['total_exs'] = self._total_exs
            self.clearml_logger.log_metrics('valid', self.parleys, valid_report)
            self.clearml_logger.flush()

        # send valid metrics to agent if the agent wants them
        if hasattr(self.agent, 'receive_metrics'):
            self.agent.receive_metrics(valid_report)

        # check which metric to look at
        new_valid = valid_report[opt['validation_metric']]

        if isinstance(new_valid, Metric):
            new_valid = new_valid.value()

        # check if this is the best validation so far
        if (
            self.best_valid is None
            or self.valid_optim * new_valid > self.valid_optim * self.best_valid
        ):
            logging.success(
                'new best {}: {:.4g}{}'.format(
                    opt['validation_metric'],
                    new_valid,
                    ' (previous best was {:.4g})'.format(self.best_valid)
                    if self.best_valid is not None
                    else '',
                )
            )
            self.best_valid = new_valid
            self.impatience = 0
            if opt.get('model_file'):
                logging.info(f"saving best valid model: {opt['model_file']}")
                self.save_model()
                self.saved = True
            if (
                opt['validation_metric_mode'] == 'max'
                and self.best_valid >= opt['validation_cutoff']
            ) or (
                opt['validation_metric_mode'] == 'min'
                and self.best_valid <= opt['validation_cutoff']
            ):
                logging.info('task solved! stopping.')
                return True
        else:
            self.impatience += 1
            logging.report(
                'did not beat best {}: {} impatience: {}'.format(
                    opt['validation_metric'], round(self.best_valid, 4), self.impatience
                )
            )
        self.validate_time.reset()

        # saving
        if opt.get('model_file') and opt.get('save_after_valid'):
            logging.info(f"saving model checkpoint: {opt['model_file']}.checkpoint")
            self.save_model('.checkpoint')

        # check if we are out of patience
        if (
            opt['validation_patience'] > 0
            and self.impatience >= opt['validation_patience']
        ):
            logging.info('ran out of patience! stopping training.')
            return True
        return False

    def _run_single_eval(
        self, opt, valid_world, max_exs, datatype, is_multitask, task, index
    ):

        # run evaluation on a single world
        valid_world.reset()

        world_logger = None
        task_opt = opt.copy()
        # set up world logger for the "test" fold
        if opt['world_logs'] and datatype == 'test':
            task_opt['world_logs'] = get_task_world_logs(
                task, opt['world_logs'], is_multitask
            )
            world_logger = WorldLogger(task_opt)

        cnt = 0
        max_cnt = max_exs if max_exs > 0 else float('inf')
        while not valid_world.epoch_done() and cnt < max_cnt:
            valid_world.parley()
            if world_logger is not None:
                world_logger.log(valid_world)
            if cnt == 0 and opt['display_examples']:
                print(valid_world.display() + '\n~~')
                print(valid_world.report())
                if opt['clearml_log'] and is_primary_worker():
                    self.clearml_logger.log_debug_samples(
                        datatype, valid_world.display(), index=index
                    )
            cnt = valid_world.report().get('exs') or 0

        if world_logger is not None:
            # dump world acts to file
            world_logger.reset()  # add final acts to logs
            if is_distributed():
                rank = get_rank()
                base_outfile, extension = os.path.splitext(task_opt['world_logs'])
                outfile = base_outfile + f'_{rank}' + extension
            else:
                outfile = task_opt['world_logs']
            world_logger.write(outfile, valid_world, file_format=opt['save_format'])

        valid_report = valid_world.report()
        if opt.get('validation_share_agent', False):
            valid_world.reset()  # make sure world doesn't remember valid data

        return valid_report

    def _run_eval(
        self,
        valid_worlds,
        opt,
        datatype,
        max_exs=-1,
        write_log=False,
        extra_log_suffix="",
    ):
        """
        Eval on validation/test data.

        :param valid_world:
            list of the pre-created validation worlds.
        :param opt:
            the options that specific the task, eval_task, etc
        :param datatype:
            the datatype to use, such as "valid" or "test"
        :param bool write_log:
            specifies to write metrics to file if the model_file is set
        :param int max_exs:
            limits the number of examples if max_exs > 0
        """

        logging.info(f'running eval: {datatype}')
        timer = Timer()
        reports = []

        max_exs_per_worker = max_exs / (len(valid_worlds) * num_workers())
        is_multitask = len(valid_worlds) > 1
        for index, v_world in enumerate(valid_worlds):
            if opt.get('evaltask'):
                task = opt['evaltask'].split(',')[index]
            else:
                task = opt['task'].split(',')[index]
            task_report = self._run_single_eval(
                opt, v_world, max_exs_per_worker, datatype, is_multitask, task, index
            )
            reports.append(task_report)

        tasks = [world.getID() for world in valid_worlds]
        named_reports = dict(zip(tasks, reports))
        report = aggregate_named_reports(
            named_reports, micro_average=self.opt.get('aggregate_micro', False)
        )
        # get the results from all workers
        report = self._sync_metrics(report)

        metrics = f'{datatype}:\n{nice_report(report)}\n'
        logging.info(f'eval completed in {timer.time():.2f}s')
        logging.report(metrics)

        # write to file
        if write_log and opt.get('model_file') and is_primary_worker():
            # Write out metrics
            with PathManager.open(
                opt['model_file'] + extra_log_suffix + '.' + datatype, 'a'
            ) as f:
                f.write(f'{metrics}\n')

        return report

    def _run_final_extra_eval(self, opt):
        final_valid_opt = copy.deepcopy(opt)
        final_valid_opt_raw = Opt.load_init(opt['final_extra_opt'])
        final_datatype = final_valid_opt_raw["datatype"]
        for k, v in final_valid_opt_raw.items():
            final_valid_opt[k] = v
        final_max_exs = (
            final_valid_opt['validation_max_exs']
            if final_valid_opt.get('short_final_eval')
            else -1
        )
        final_valid_world = load_eval_worlds(
            self.agent, final_valid_opt, final_datatype
        )
        final_valid_report = self._run_eval(
            final_valid_world,
            final_valid_opt,
            final_datatype,
            final_max_exs,
            write_log=True,
            extra_log_suffix="_extra",
        )
        if opt['wandb_log'] and is_primary_worker():
            self.wb_logger.log_final(final_datatype, final_valid_report)
        if opt['clearml_log'] and is_primary_worker():
            self.clearml_logger.log_final(final_datatype, final_valid_report)

        return final_valid_report

    def _sync_metrics(self, metrics):
        """
        Sync training metrics across workers.

        A handful of special cases are handled as exceptions, and the remaining metrics
        are simply averaged across workers.
        """
        if not is_distributed():
            # nothing special needed
            return metrics
        all_versions = all_gather_list(metrics)
        return aggregate_unnamed_reports(all_versions)

    def _compute_eta(
        self, epochs_completed: float, time_elapsed: float, steps_taken: int
    ):
        """
        Compute the estimated seconds remaining in training.

        :param float epochs_completed: number of epochs already completed.
        :param float time_elapsed: total time spent already, in seconds.
        :return: ETA in seconds, or None if not computable
        """
        # start off with no estimate
        eta = None

        # Determine time_left and num_epochs
        max_epochs = self.opt.get('num_epochs', 0)
        if max_epochs > 0 and epochs_completed > 0:
            epoch_progress = epochs_completed / max_epochs
            eta = (1 - epoch_progress) * time_elapsed / epoch_progress

        max_training_time = self.opt.get('max_training_time', -1)
        if max_training_time > 0:
            time_left = max_training_time - time_elapsed
            if eta is None or time_left < eta:
                eta = time_left

        max_train_steps = self.opt.get('max_train_steps', -1)
        if max_train_steps > 0 and steps_taken > 0:
            steps_progress = steps_taken / max_train_steps
            eta = (1 - steps_progress) * time_elapsed / steps_progress

        return eta

    def _get_time(self, world: World) -> Tuple[float, float, float]:
        """
        Return train, log, and validate timing.

        If relying on the time for validation/logging/max train time purposes,
        we sync and return primary worker's time.

        Otherwise, it's not super relevant what we do here.

        **SIDE EFFECT**: Update _total_epochs trained.

        :param world:
            current running world

        :return (train, log, valid):
            return time for each of train, log, and validation
        """
        if (
            self.max_train_time < float('inf')
            or self.log_every_n_secs < float('inf')
            or self.val_every_n_secs < float('inf')
            or self.val_every_n_epochs < float('inf')
            or self.max_num_epochs < float('inf')
        ):
            self._total_epochs = self._preempted_epochs + sum(
                all_gather_list(world.get_total_epochs())
            )
            train_time, log_time, validate_time, save_time = sync_object(
                (
                    self.train_time.time(),
                    self.log_time.time(),
                    self.validate_time.time(),
                    self.save_time.time(),
                )
            )
        else:
            train_time, log_time, validate_time, save_time = (
                self.train_time.time(),
                self.log_time.time(),
                self.validate_time.time(),
                self.save_time.time(),
            )
            self._total_epochs = self._preempted_epochs + (
                num_workers() * world.get_total_epochs()
            )

        return train_time, log_time, validate_time, save_time

    def log(self):
        """
        Output a training log entry.
        """
        opt = self.opt
        if opt['display_examples']:
            print(self.world.display() + '\n~~')
        logs = []
        # get report
        train_report = self.world.report()
        train_report = self._sync_metrics(train_report)
        self.world.reset_metrics()

        train_report_trainstats = dict_report(train_report)
        train_report_trainstats['total_epochs'] = self._total_epochs
        train_report_trainstats['total_exs'] = self._total_exs
        train_report_trainstats['parleys'] = self.parleys
        train_report_trainstats['train_steps'] = self._train_steps
        train_report_trainstats['train_time'] = self.train_time.time()
        self.train_reports.append(train_report_trainstats)

        # time elapsed
        logs.append(f'time:{self.train_time.time():.0f}s')
        logs.append(f'total_exs:{self._total_exs}')
        logs.append(f'total_steps:{self._train_steps}')

        if self._total_epochs >= 0:
            # only if it's unbounded
            logs.append(f'epochs:{self._total_epochs:.2f}')

        time_left = self._compute_eta(
            self._total_epochs, self.train_time.time(), self._train_steps
        )
        if time_left is not None:
            logs.append(f'time_left:{max(0,time_left):.0f}s')

        log = '{}\n{}\n'.format(' '.join(logs), nice_report(train_report))
        logging.info(log)
        self.log_time.reset()
        self._last_log_steps = 0

        if opt['tensorboard_log'] and is_primary_worker():
            self.tb_logger.log_metrics('train', self.parleys, train_report)
        if opt['wandb_log'] and is_primary_worker():
            self.wb_logger.log_metrics('train', self.parleys, train_report)
        if opt['clearml_log'] and is_primary_worker():
            self.clearml_logger.log_metrics('train', self.parleys, train_report)

        return train_report

    def train_steps(self):
        """
        Core training loop.

        Yields a metrics dict with each log.
        """
        logging.info('training...')
        opt = self.opt
        world = self.world
        with world:
            while True:
                # do one example / batch of examples
                try:
                    world.parley()
                except StopTrainException as e:
                    logging.info(f"Stopping from {e}")
                    break

                self.parleys += 1
                self._train_steps = self.parleys // self.update_freq
                self._last_log_steps += 1 / self.update_freq

                # the following additionally updates self._total_epochs
                train_time, log_time, validate_time, save_time = self._get_time(world)
                # get the total training examples done, compute epochs
                exs_per_epoch = world.num_examples()
                self._total_exs = int(np.round(self._total_epochs * exs_per_epoch))

                # check counters and timers
                if self._total_epochs >= self.max_num_epochs:
                    yield self.log()
                    logging.info(
                        f'num_epochs completed:{self.max_num_epochs} time elapsed:{train_time}s'
                    )
                    break
                if train_time > self.max_train_time:
                    logging.info(f'max_train_time elapsed:{train_time}s')
                    break
                if self._train_steps >= self.max_train_steps:
                    logging.info(
                        f'max_train_steps elapsed:{self._train_steps} '
                        f'time elapsed:{train_time}s'
                    )
                    break
                if (
                    log_time > self.log_every_n_secs
                    or self._last_log_steps >= self.log_every_n_steps
                ):
                    yield self.log()
                if (
                    validate_time > self.val_every_n_secs
                    or self._total_epochs - self.last_valid_epoch
                    >= self.val_every_n_epochs
                    or self._train_steps - self._last_valid_steps
                    >= self.val_every_n_steps
                ):
                    try:
                        # log before we validate
                        if self._last_log_steps:
                            yield self.log()
                        world.reset_metrics()
                        stop_training = self.validate()
                    except StopTrainException:
                        break
                    # reset the log time because we logged right before validating
                    self.log_time.reset()
                    self.last_valid_epoch = self._total_epochs
                    self._last_valid_steps = self._train_steps
                    if stop_training:
                        break
                    # make sure metrics are clean before we log
                    world.reset_metrics()
                if save_time > self.save_every_n_secs and opt.get('model_file'):
                    logging.info(
                        f"saving model checkpoint: {opt['model_file']}.checkpoint"
                    )
                    if opt['tensorboard_log'] and is_primary_worker():
                        self.tb_logger.flush()
                    self.save_model('.checkpoint')
                    self.save_time.reset()

        if not sync_object(self.saved):
            # save agent
            self.save_model()

        # there's a rare edge case where the we never saved the model, and we try
        # # to reload it. This sync_object ensures all workers wait for the primary
        # worker to finish flushing before loading from disk.
        sync_object(None)
        if opt.get('model_file'):
            # clean up all our memory, just to make sure we don't OOM on GPU when
            # reloading the world
            del world
            del self.world
            del self.agent
            del self.valid_worlds
            # reload best validation model
            self.agent = create_agent(opt)

    def train(self):
        """
        Perform a training run.

        :return: tuple of reports (validation_report, test_report)
        """
        opt = self.opt
        for _train_log in self.train_steps():
            # we've already done what we need in these
            pass

        # perform final validation/testing
        valid_worlds = load_eval_worlds(self.agent, opt, 'valid')
        max_exs = opt['validation_max_exs'] if opt.get('short_final_eval') else -1
        self.final_valid_report = self._run_eval(
            valid_worlds, opt, 'valid', max_exs, write_log=True
        )
        test_worlds = load_eval_worlds(self.agent, opt, 'test')
        self.final_test_report = self._run_eval(
            test_worlds, opt, 'test', max_exs, write_log=True
        )

        if opt['wandb_log'] and is_primary_worker():
            self.wb_logger.log_final('valid', self.final_valid_report)
            self.wb_logger.log_final('test', self.final_test_report)
            self.wb_logger.finish()
        if opt['clearml_log'] and is_primary_worker():
            self.clearml_logger.log_final('Validation Report', self.final_valid_report)
            self.clearml_logger.log_final('Test Report', self.final_test_report)

        if valid_worlds:
            for valid_world in valid_worlds:
                valid_world.shutdown()
        if test_worlds:
            for test_world in test_worlds:
                test_world.shutdown()

        print_announcements(opt)

        if opt['final_extra_opt'] != '':
            self.final_extra_valid_report = self._run_final_extra_eval(opt)

        if opt['wandb_log'] and is_primary_worker():
            self.wb_logger.finish()

        if opt['clearml_log'] and is_primary_worker():
            self.clearml_logger.upload_artifact('dictionary', opt['dict_file'])
            self.clearml_logger.close()

        self._save_train_stats()

        return self.final_valid_report, self.final_test_report


@register_script('train_model', aliases=['tm', 'train'])
class TrainModel(ParlaiScript):
    @classmethod
    def setup_args(cls):
        return setup_args()

    def run(self):
        self.train_loop = TrainLoop(self.opt)
        if self.opt['seed'] is not None:
            set_seed(self.opt['seed'])
        return self.train_loop.train()


if __name__ == '__main__':
    TrainModel.main()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from abc import ABC, abstractmethod
import copy
from typing import List, Optional, Union

from parlai.core.agents import Agent
from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.teachers import MultiTaskTeacher
from parlai.tasks.convai2.agents import NormalizedTeacher
from parlai.tasks.fits.agents import FitsBaseTeacher
from parlai.tasks.light_dialog.agents import DefaultTeacher as LightTeacher
from parlai.tasks.light_dialog_wild.agents import DefaultTeacher as LightWildTeacher
from parlai.tasks.msc.agents import (
    SessionBaseMscTeacher,
    SessionBasePersonaSummaryTeacher,
)
from parlai.tasks.saferdialogues.agents import SaferDialoguesTeacher
from parlai.tasks.taskmaster2.agents import Taskmaster2Parser

#########
# Mixin #
#########


class BB3TeacherMixin(Agent):
    """
    This mixin is to ensure that messages from the teachers contain the proper fields.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if hasattr(self, 'tasks'):
            for t in self.tasks:
                t.id = self._id

    def get(self, episode_idx, entry_idx=None):
        example = super().get(episode_idx, entry_idx)
        if isinstance(example, Message) and 'id' in example:
            example.force_set('id', self._id)
        else:
            example['id'] = self._id
        return example

    @property
    def id(self) -> str:
        return self._id

    @id.setter
    def id(self, str):
        self._id = type(self).__name__


##############
# Multitasks #
##############


class BB3MultitaskTeacher(MultiTaskTeacher, ABC):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        MultiTaskTeacher.add_cmdline_args(parser, partial_opt)
        parser.add_argument(
            '--blenderbot3-base-agent',
            type=str,
            default='r2c2',
            choices=['r2c2', 'opt'],
            help='Base agent for which to format tasks.',
        )
        return parser

    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        tasks = [
            f"projects.bb3.tasks.{opt['blenderbot3_base_agent']}_{self.get_task_type()}_tasks:{teacher}"
            for teacher in self.get_teachers()
        ]
        opt['task'] = ','.join(tasks)
        opt['multitask_weights'] = self.get_multitask_weights()
        super().__init__(opt, shared)

    @abstractmethod
    def get_task_type(self) -> str:
        """
        Return the task type.
        """

    @abstractmethod
    def get_teachers(self) -> List[str]:
        """
        Get the teachers for multitask.
        """

    @abstractmethod
    def get_multitask_weights(self) -> Union[List[int], str]:
        """
        Get the weights for multitask.
        """


class AlwaysSearchTeacher(BB3MultitaskTeacher):
    def get_teachers(self) -> List[str]:
        return [
            'NQOpenSearchDecisionTeacher',
            'SquadSearchDecisionTeacher',
            'TriviaQASearchDecisionTeacher',
        ]

    def get_task_type(self) -> str:
        return 'decision'

    def get_multitask_weights(self) -> Union[List[int], str]:
        return [1] * len(self.get_teachers())


class MaybeSearchTeacher(BB3MultitaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        NormalizedTeacher.add_cmdline_args(parser, partial_opt)
        SessionBaseMscTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def get_teachers(self) -> List[str]:
        return [
            'WowSearchDecisionTeacher',
            'WoiSearchDecisionTeacher',
            'Convai2SearchDecisionTeacher',
            'EDSearchDecisionTeacher',
            'MSCSearchDecisionTeacher',
        ]

    def get_multitask_weights(self) -> Union[List[int], str]:
        return [1] * len(self.get_teachers())

    def get_task_type(self) -> str:
        return 'decision'


class MemoryDecisionTeacher(BB3MultitaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        NormalizedTeacher.add_cmdline_args(parser, partial_opt)
        SessionBaseMscTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def get_teachers(self) -> List[str]:
        return [
            'Convai2MemoryDecisionTeacher',
            'MSCMemoryDecisionTeacher',
            'BSTMemoryDecisionTeacher',
            'EDMemoryDecisionTeacher',
        ]

    def get_multitask_weights(self) -> Union[List[int], str]:
        return [3, 3, 1, 1]

    def get_task_type(self) -> str:
        return 'decision'


class SearchQueryGenerationTeacher(BB3MultitaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        FitsBaseTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def get_teachers(self) -> List[str]:
        return ['WoiSearchQueryTeacher', 'FitsSearchQueryTeacher']

    def get_multitask_weights(self) -> Union[List[int], str]:
        return [1, 1]

    def get_task_type(self) -> str:
        return 'search_generation'


class MemoryGenerationTeacher(BB3MultitaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        SessionBasePersonaSummaryTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def get_teachers(self) -> List[str]:
        return ['MscMemoryGenerationTeacher']

    def get_multitask_weights(self) -> Union[List[int], str]:
        return [1]

    def get_task_type(self) -> str:
        return 'memory_generation'


class MemoryKnowledgeGenerationTeacher(BB3MultitaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        NormalizedTeacher.add_cmdline_args(parser, partial_opt)
        SessionBaseMscTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def get_teachers(self) -> List[str]:
        return [
            'BSTMemoryKnowledgePersOverlapTeacher',
            'BSTMemoryKnowledgeUttOverlapTeacher',
            'Convai2MemoryKnowledgePersOverlapTeacher',
            'Convai2MemoryKnowledgeUttOverlapTeacher',
            'EDMemoryKnowledgePersOverlapTeacher',
            'EDMemoryKnowledgeUttOverlapTeacher',
            'MSCMemoryKnowledgePersOverlapTeacher',
            'MSCMemoryKnowledgeUttOverlapTeacher',
        ]

    def get_multitask_weights(self) -> Union[List[int], str]:
        """
        Weighting justification:

        6 = persona overlap for Convai2/MSC
        3 = utterance overlap for Convai2/MSC
        2 = persona overlap for BST/ED
        1 = utterance overlap for BST/ED
        """
        return [2, 1, 6, 3, 2, 1, 6, 3]

    def get_task_type(self) -> str:
        return 'knowledge'


class SearchKnowledgeGenerationTeacher(BB3MultitaskTeacher):
    def get_teachers(self) -> List[str]:
        return [
            'MsMarcoSearchKnowledgeTeacher',
            'NQSearchKnowledgeTeacher',
            'NQOpenSearchKnowledgeTeacher',
            'NQOpenDialoguesSearchKnowledgeTeacher',
            'SquadSearchKnowledgeTeacher',
            'TriviaQASearchKnowledgeTeacher',
            'WowSearchKnowledgeTeacher',
            'WoiSearchKnowledgeTeacher',
            'FitsSearchKnowledgeTeacher',
        ]

    def get_multitask_weights(self) -> Union[List[int], str]:
        return [1] * len(self.get_teachers())

    def get_task_type(self) -> str:
        return 'knowledge'


class EntityKnowledgeGenerationTeacher(BB3MultitaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        NormalizedTeacher.add_cmdline_args(parser, partial_opt)
        SessionBaseMscTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def get_teachers(self) -> List[str]:
        return [
            'BSTEntityKnowledgeTeacher',
            'Convai2EntityKnowledgeTeacher',
            'EDEntityKnowledgeTeacher',
            'MSCEntityKnowledgeTeacher',
        ]

    def get_multitask_weights(self) -> Union[List[int], str]:
        return [1] * len(self.get_teachers())

    def get_task_type(self) -> str:
        return 'knowledge'


class SearchDialogueGenerationTeacher(BB3MultitaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        Taskmaster2Parser.add_cmdline_args(parser, partial_opt)
        FitsBaseTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def get_teachers(self) -> List[str]:
        return [
            'MsMarcoSearchDialogueTeacher',
            'WowSearchDialogueTeacher',
            'WoiSearchDialogueTeacher',
            'GoogleSgdSearchDialogueTeacher',
            'TaskmasterSearchDialogueTeacher',
            'Taskmaster2SearchDialogueTeacher',
            'Taskmaster3SearchDialogueTeacher',
            'FitsSearchDialogueTeacher',
            'FunpediaSearchDialogueTeacher',
        ]

    def get_multitask_weights(self) -> Union[List[int], str]:
        return [2] * (len(self.get_teachers()) - 1) + [1]

    def get_task_type(self) -> str:
        return 'dialogue'


class EntityDialogueGenerationTeacher(BB3MultitaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        NormalizedTeacher.add_cmdline_args(parser, partial_opt)
        SessionBaseMscTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def get_teachers(self) -> List[str]:
        return [
            'BSTEntityDialogueTeacher',
            'Convai2EntityDialogueTeacher',
            'EDEntityDialogueTeacher',
            'MSCEntityDialogueTeacher',
        ]

    def get_multitask_weights(self) -> Union[List[int], str]:
        """
        Justification for weights:

        Convai2 and MSC are quite larger compared to BST and ED in this task.

        Rather than have stochastic (which is like 10:1), let's just make it 4:1
        """
        return [1, 4, 1, 4]

    def get_task_type(self) -> str:
        return 'dialogue'


class MemoryDialogueGenerationTeacher(BB3MultitaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        NormalizedTeacher.add_cmdline_args(parser, partial_opt)
        SessionBaseMscTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def get_teachers(self) -> List[str]:
        return [
            'BSTMemoryDialogueFromPersOverlapTeacher',
            'BSTMemoryDialogueFromUttOverlapTeacher',
            'Convai2MemoryDialogueFromPersOverlapTeacher',
            'Convai2MemoryDialogueFromUttOverlapTeacher',
            'EDMemoryDialogueFromPersOverlapTeacher',
            'EDMemoryDialogueFromUttOverlapTeacher',
            'MSCMemoryDialogueFromPersOverlapTeacher',
            'MSCMemoryDialogueFromUttOverlapTeacher',
        ]

    def get_multitask_weights(self) -> Union[List[int], str]:
        """
        Justification:

        Same as Memory Knowledge Generation Teacher
        """
        return [2, 1, 6, 3, 2, 1, 6, 3]

    def get_task_type(self) -> str:
        return 'dialogue'


class VanillaDialogueGenerationTeacher(BB3MultitaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        NormalizedTeacher.add_cmdline_args(parser, partial_opt)
        SessionBaseMscTeacher.add_cmdline_args(parser, partial_opt)
        SaferDialoguesTeacher.add_cmdline_args(parser, partial_opt)
        LightTeacher.add_cmdline_args(parser, partial_opt)
        LightWildTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def get_teachers(self) -> List[str]:
        return [
            'WowVanillaDialogueTeacher',
            'WoiVanillaDialogueTeacher',
            'Convai2VanillaDialogueTeacher',
            'EDVanillaDialogueTeacher',
            'MSCVanillaDialogueTeacher',
            'SaferdialoguesVanillaDialogueTeacher',
            'LightVanillaDialogueTeacher',
            'LightWildVanillaDialogueTeacher',
            'BSTStyleGroundingDialogueTeacher',
            'Convai2StyleGroundingDialogueTeacher',
        ]

    def get_multitask_weights(self) -> Union[List[int], str]:
        """
        Justification:

        All equal
        """
        return [1] * len(self.get_teachers())

    def get_task_type(self) -> str:
        return 'dialogue'



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.opt import Opt
from parlai.core.message import Message
from parlai.core.params import ParlaiParser
from parlai.core.teachers import DialogTeacher
from tqdm import tqdm
from typing import Any, Dict, Iterable, List, Optional, Tuple

"""
Base classes for dealing with datasets that represent Reasoning

We do this because reasoning datasets oftentimes have the same structure and it's obnoxious
having to deal with some of the nitty logistical details (ex. special tokens) over and over again.

Also makes it easy to have a common sent of synthetic pertubations for reasoning steps.
"""

# Couple typedefs to make it clearer what we're doing
t_QUESTION = str
t_ANSWER_PREFIX_TOKEN = str
t_ANSWER = str

t_REASON_PREFIX_TOKEN = str
t_REASON = str
t_REASON_PROCESSED_RESULT = Tuple[t_REASON_PREFIX_TOKEN, t_REASON, Dict]


class AbstractQuestionAnswer:
    """
    Used inside of an `AbstractReasoningTeacher` to define how the "Question" and
    "Answer" part of an example should be serialized out to a string, given a dict
    object that includes all necessary information for the example.

    The Question includes all context (ex. input background) necessary to generate the
    Answer. This class is responsible for handling its own tokenization.
    """

    def __init__(self, opt: Opt):
        self.opt = opt
        self.inner_separator = opt.get("inner_separator_token", "\t")
        self.answer_token = opt.get("answer_token", "ANSWER: ")

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        group = parser.add_argument_group("Abstract Question Answer args")
        group.add_argument(
            "--inner-separator-token",
            type=str,
            default="\t",
            help="See abstract reasoning teacher help for usage. Including here as well for consistency + testing purposes",
        )
        group.add_argument(
            "--answer-token",
            type=str,
            default="ANSWER: ",
            help="Token in front of the answer. Child classes can set the default to something else if desired",
        )
        return parser

    def get_question_answer(
        self, example_dict: Dict
    ) -> Tuple[t_QUESTION, t_ANSWER_PREFIX_TOKEN, t_ANSWER, Dict]:
        """
        Given a dict object that includes all of the information necessary for an
        example, serialize Question and Answer strings that will be used to construct
        the `text` object in a ParlAI Teacher Message. Called from `setup_data()` in
        `AbstractReasoningTeacher`.

        For example, the 'question text' for a multiple choice version of this class may include different
        presentation of prefixes to different answer choices (ex. a,b,c vs 1,2,3). Depending on external flags,
        the `answer prefix token` may be a straightforward "ANSWER:" or it may be something like "CHOICE:";
        the `answer text` itself may be the answer itself or the prefix, etc.

        Output is:

            str - question text (Including any structured prefix token)
            str - answer prefix token
            str - answer text
            Dict - any metadata to append to the final Message object

        We return the `answer prefix token` as a string out of this function in case we want to prompt the
        message with it as the last line of the input, rather than as the first part of the label.
        """
        raise NotImplementedError("Derived class must implement")


class AbstractReason:
    """
    Used inside of an `AbstractReasoningTeacher` to define how the "Reason" part of an
    example should be serialized out to a string, given a dict object that includes all
    necessary information for the example.

    Downstream classes can also implement `perturbations` on these reasons. (See
    `reason_types/step_by_step.py` for examples of this.)
    """

    def __init__(self, opt: Opt):
        self.opt = opt
        self.inner_separator = opt.get("inner_separator_token", "\t")

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        group = parser.add_argument_group("Abstract Reason args")
        group.add_argument(
            "--inner-separator-token",
            type=str,
            default="\t",
            help="See abstract reasoning teacher help for usage. Including here as well for consistency + testing purposes",
        )
        return parser

    def get_full_reason_text(self, example_dict: Dict) -> t_REASON_PROCESSED_RESULT:
        """
        Given a dict object that includes all of the information necessary for an
        example, serialize Reason strings that will be used to construct the `text`
        object in a ParlAI Teacher Message. Called from `setup_data()` in
        `AbstractReasoningTeacher`.

        Perturbations to reasoning chains should be implemented in this function.

        Output is:

            str - reason prefix token
            str - reason string
            Dict - any metadata to append to the final Message object

        We return the `reason prefix token` as a string out of this class in case we want to
        prompt the message with it as the last line of the input, rather than as the first part of the label.
        """

        raise NotImplementedError("Derived class must implement")

    def batch_get_full_reason_text(
        self, examples: List[Dict]
    ) -> List[t_REASON_PROCESSED_RESULT]:
        """
        Can be overridden to parallelize expensive parts of the get_full_reason_text
        operation.

        Only called when --preload-batch-size is set to a value greater than one.
        """
        return [self.get_full_reason_text(e) for e in examples]


class AbstractReasoningTeacher(DialogTeacher):
    """
    Teacher that builds up Message objects based on child `question and answer` and
    `reason` classes.

    Also handles making exemplars.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        group = parser.add_argument_group("Abstract Reasoning Teacher args")
        #####
        # Token options
        #####
        group.add_argument(
            "--inner-separator-token",
            type=str,
            default="\t",
            help="Token that we use to separate components within an example. Intentially the same"
            + "as in AbstractReason and AbstractQuestionAnswer",
        )
        group.add_argument(
            "--example-separator-token",
            type=str,
            default="\n",
            help="Token that we use to separate different examples. Set to '\\n' as a default since"
            + "some agents (ex. GPTZ) will use '\\n' as a way to separate turns",
        )
        #####
        # Options for combining/including reasons with questions/answers
        #####
        group.add_argument(
            "--include-reason",
            type=bool,
            default=True,
            help="Should we include the reason as part of the generation/exemplars?",
        )
        group.add_argument(
            "--reason-before-answer",
            type=bool,
            default=True,
            help="When we include a reason, should it come before or after the answer? Useful for"
            + "designating pre-hoc vs post-hoc justification",
        )
        group.add_argument(
            "--initial-generation-target",
            type=str,
            choices=["reason", "answer", "other"],
            default="reason",
            help="If we are not including the reason, this does nothing and goes with assuming `answer`"
            + "is the target since that is the only option. This, in conjunction with `--reason-before-answer`"
            + "determines what is included in the input vs the output; we will construct the input/output"
            + "appropriately based on these two flags if `reason` or `answer` is selected. If 'other' is selected,"
            + "`self.make_custom_input_text_label()` will be called. (Last of these most useful for if we"
            + "want to do say, step-by-step reasoning",
        )
        group.add_argument(
            "--include-generation-target-token-in-input",
            type=bool,
            default=True,
            help="Sometimes we want to include the token representing the reason or the answer at the end"
            + "of our input, since this can help prime the model to do the correct generation",
        )
        #####
        # Options for task prompts + exemplars
        #####
        group.add_argument(
            "--use-task-prompt",
            type=bool,
            default=True,
            help="Use the prompt from `get_task_prompt()` before all reasons, questions, and answers",
        )
        group.add_argument(
            "--task-prompt",
            type=str,
            default="",
            help="`get_task_prompt()` looks at this by default. Child classes can override this value.",
        )
        group.add_argument(
            "--exemplar-idx",
            type=int,
            nargs="*",
            default=[],
            help="IDX of exemplars to use. Looks data up from a `self.exemplars_raw` field that should be"
            + "set in `__init__()`. NOTE: Including this in the base class as a convenience; individual"
            + "datasets may find it useful to have their own `--<dataset name>-exemplar-idx` arguments for"
            + "multitasking, since this will be shared across tasks",
        )
        #####
        # Data loading
        #####
        group.add_argument(
            "--preload-batch-size",
            type=int,
            default=1,
            help=(
                "If greater than 1, data will be preloaded in batches of this size. This is useful when "
                "data loading involves an expensive operation, e.g. paraphrase perturbation."
            ),
        )
        group.add_argument(
            "--skip-perturbation-failures",
            type=bool,
            default=False,
            help=(
                "Some chains can not be perturbes for certain perturbation types. Set this flag to return "
                "only perturbed chains and drop non-perturbed examples"
            ),
        )

        parser = cls.get_reason_class().add_cmdline_args(parser, partial_opt)
        parser = cls.get_question_answer_class().add_cmdline_args(parser, partial_opt)

        return parser

    @classmethod
    def get_reason_class(self) -> AbstractReason:
        raise NotImplementedError("Derived class must implement")

    @classmethod
    def get_question_answer_class(self) -> AbstractQuestionAnswer:
        raise NotImplementedError("Derived class must implement")

    def get_task_prompt(self) -> str:
        """
        To keep flexibility, do not override this function.
        """
        prompt = self.opt.get("task_prompt", "")
        if prompt == "":
            raise RuntimeError(
                "Got an empty string for the prompt - is this intentional? If not, set with "
                "`parser.set_defaults(task_prompt='<INSERT PROMPT>')` in `add_cmdline_args`."
            )
        return prompt

    def __init__(self, opt, shared=None):
        self.reason_instance = self.get_reason_class()(opt)
        self.question_answer_instance = self.get_question_answer_class()(opt)
        super().__init__(opt, shared)
        self.paraphrase_cache = {}

    def get_data_for_fold(self, fold):
        raise NotImplementedError("Derived class must implement")

    def make_custom_input_text_label(self):
        raise NotImplementedError("Derived class to (optionally) implement")

    def get_exemplar(self, idx):
        return self.exemplars_raw[idx]

    def setup_data(self, fold) -> Iterable[Tuple[Message, bool]]:
        if self.opt.get('preload_batch_size', 1) > 1:
            if not hasattr(self, '_preloaded_data'):
                self._preloaded_data = self._preload_data(fold)
            data_iterator = self._preloaded_data
        else:
            data_iterator = self._data_generator(fold)
        for message in data_iterator:
            yield message, True

    def _data_generator(self, fold) -> Iterable[Message]:
        example_separator = self.opt.get("example_separator_token")
        inner_separator = self.opt.get("inner_separator_token")
        skip_perturbation_failures = self.opt.get("skip_perturbation_failures")

        prompt_and_examplars = self._build_prompt_and_exemplars(
            inner_separator,
            example_separator,
        )

        #### Set up individual examples
        for example in self.get_data_for_fold(fold):
            reason_tuple = None
            if self.opt.get("include_reason", True):
                try:
                    reason_tuple = self.reason_instance.get_full_reason_text(example)
                except (RuntimeError, IndexError) as e:
                    if skip_perturbation_failures:
                        print(f"WARNING: {e}")
                        continue
                    else:
                        raise e
            message = self._message_from_example(
                example=example,
                prompt_and_examplars=prompt_and_examplars,
                inner_separator=inner_separator,
                reason_tuple=reason_tuple,
            )
            yield message

    def _preload_data(self, fold) -> List[Message]:
        example_separator = self.opt.get("example_separator_token")
        inner_separator = self.opt.get("inner_separator_token")

        prompt_and_examplars = self._build_prompt_and_exemplars(
            inner_separator,
            example_separator,
        )

        all_messages = []
        try:
            print("Pre-loading data...")
            batch = []
            for example in tqdm(list(self.get_data_for_fold(fold))):
                batch.append(example)
                if len(batch) >= self.opt['preload_batch_size']:
                    all_messages.extend(
                        self._load_data_batch(
                            batch,
                            prompt_and_examplars,
                            inner_separator,
                        )
                    )
                    batch = []
            if len(batch) > 0:
                all_messages.extend(
                    self._load_data_batch(
                        batch,
                        prompt_and_examplars,
                        inner_separator,
                    )
                )
        finally:
            self.reason_instance.save_cache()

        return all_messages

    def _load_data_batch(
        self,
        example_batch: List[Dict],
        prompt_and_examplars: str,
        inner_separator: str,
    ) -> List[Message]:
        full_reason_texts = [None] * len(example_batch)
        if self.opt.get("include_reason", True):
            full_reason_texts = self.reason_instance.batch_get_full_reason_text(
                example_batch
            )
        message_list = [
            self._message_from_example(
                example=example,
                prompt_and_examplars=prompt_and_examplars,
                inner_separator=inner_separator,
                reason_tuple=full_reason,
            )
            for example, full_reason in zip(example_batch, full_reason_texts)
        ]
        return message_list

    def _build_prompt_and_exemplars(
        self, inner_separator: str, example_separator: str
    ) -> str:
        #### Set up our prompt and any exemplars we would like to use for the whole dataset
        prompt_and_examplars_raw = []
        if self.opt.get("use_task_prompt", True):
            prompt_and_examplars_raw.append(self.get_task_prompt())
        for idx in self.opt.get("exemplar_idx", []):
            (
                question,
                answer_prefix,
                answer,
                _,
            ) = self.question_answer_instance.get_question_answer(
                self.get_exemplar(idx)
            )
            exemplar = f"{question}{inner_separator}"
            if self.opt.get("include_reason", True):
                reason_prefix, reason, _ = self.reason_instance.get_full_reason_text(
                    self.get_exemplar(idx)
                )
                if self.opt.get("reason_before_answer", True):
                    exemplar += f"{reason_prefix}{reason}{inner_separator}{answer_prefix}{answer}"
                else:
                    exemplar += f"{answer_prefix}{answer}{inner_separator}{reason_prefix}{reason}"
            else:
                exemplar += f"{answer_prefix}{answer}"
            prompt_and_examplars_raw.append(exemplar)
        prompt_and_examplars = example_separator.join(prompt_and_examplars_raw)
        if len(prompt_and_examplars) > 0:
            prompt_and_examplars += example_separator
        return prompt_and_examplars

    def _message_from_example(
        self,
        example: Dict[str, Any],
        prompt_and_examplars: str,
        inner_separator: str,
        reason_tuple: Optional[t_REASON_PROCESSED_RESULT] = None,
    ) -> Message:
        (
            question,
            answer_prefix,
            answer,
            qa_dict,
        ) = self.question_answer_instance.get_question_answer(example)

        message_dict = qa_dict

        if self.opt.get("include_reason", True):
            assert reason_tuple is not None
            (
                reason_prefix,
                reason,
                reason_dict,
            ) = reason_tuple

            message_dict.update(reason_dict)

            if self.opt.get("initial_generation_target", "reason") == "other":
                (
                    example_text,
                    label_prefix,
                    label,
                ) = self.make_custom_input_text_label(
                    question,
                    answer_prefix,
                    answer,
                    reason_prefix,
                    reason,
                    message_dict,
                )
            else:
                example_text = f"{question}{inner_separator}"
                label_prefix = ""
                label = ""
                if self.opt.get("reason_before_answer", True):
                    if self.opt.get("initial_generation_target", "reason") == "reason":
                        label_prefix = reason_prefix
                        label = f"{reason}{inner_separator}{answer_prefix}{answer}"
                    else:
                        example_text += f"{reason_prefix}{reason}{inner_separator}"
                        label_prefix = answer_prefix
                        label = answer
                else:  # answer before reason
                    if self.opt.get("initial_generation_target", "reason") == "answer":
                        label_prefix = answer_prefix
                        label = f"{answer}{inner_separator}{reason_prefix}{reason}"
                    else:
                        example_text += f"{answer_prefix}{answer}{inner_separator}"
                        label_prefix = reason_prefix
                        label = reason

        else:  # case without reason is simple, just get the answer
            example_text = f"{question}{inner_separator}"
            label_prefix = answer_prefix
            label = answer

        if self.opt.get("include_generation_target_token_in_input", True):
            example_text += label_prefix
        else:
            label = label_prefix + label

        message_dict["text"] = prompt_and_examplars + example_text
        message_dict["label"] = label

        return Message(message_dict)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Taskmaster-3 implementation for ParlAI.

No official train/valid/test splits are available as of 2020-05-18, so we make our own
splits.
"""

from parlai.core.params import ParlaiParser
import os
import pandas as pd
from parlai.core.opt import Opt
import parlai.core.tod.tod_core as tod
import json
from typing import Optional
from parlai.utils.data import DatatypeHelper
from parlai.utils.io import PathManager

import parlai.tasks.taskmaster3.build as build_
import parlai.core.tod.tod_agents as tod_agents

SILENCE_TOKEN = "__SILENCE__"


class Taskmaster3Parser(tod_agents.TodStructuredDataParser):
    """
    Abstract data loader.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        return super().add_cmdline_args(parser, partial_opt)

    def __init__(self, opt: Opt, shared=None):
        self.fold = DatatypeHelper.fold(opt["datatype"])
        opt["datafile"] = self.fold
        self.dpath = os.path.join(opt["datapath"], "taskmaster-3")
        build_.build(opt)
        super().__init__(opt, shared)

    def _load_data(self, fold):
        # load up the ontology
        fn = os.path.join(self.dpath, "apis.json")
        with PathManager.open(fn, "r") as f:
            api_schema_raw = json.load(f)

        chunks = []
        for section in range(20):
            section = f"{section:02d}"
            with PathManager.open(
                os.path.join(self.dpath, f"data_{section}.json")
            ) as f:
                subset = pd.read_json(f)
            chunks.append(subset)
        chunks = pd.concat(chunks, axis=0)
        # deterministic shuffle data for splits
        chunks = chunks.sample(frac=1.0, random_state=42)
        split_size = len(chunks) // 10
        if fold == "train":
            chunks = chunks[: split_size * 8]
        elif fold == "valid":
            chunks = chunks[split_size * 8 : split_size * 9]
        elif fold == "test":
            chunks = chunks[split_size * 9 :]
        return chunks, api_schema_raw

    def _parse_apis_to_slots(self, apis_blob):
        calls = []
        responses = []
        for api in apis_blob:
            call = api["args"]
            call[tod.STANDARD_API_NAME_SLOT] = api["name"]
            calls.append(call)
            response = api["response"]
            response[tod.STANDARD_API_NAME_SLOT] = api["name"]
            responses.append(response)
        return calls, responses

    def _get_utterance_and_apis_for_speaker(self, speaker, utterances, idx):
        utts = []
        calls = []
        responses = []
        while idx < len(utterances):
            here = utterances[idx]
            if here["speaker"] != speaker:
                break
            utts.append(here["text"])
            here_calls, here_responses = self._parse_apis_to_slots(here.get("apis", []))
            calls += here_calls
            responses += here_responses
            idx += 1
        return idx, "\n".join(utts), calls, responses

    def _parse_to_api_schema(self, raw):
        result = []
        for key, val in raw.items():
            here = {}
            here[tod.STANDARD_API_NAME_SLOT] = key
            req = val.get("args", {}).get("all_of", [])
            opt = val.get("args", {}).get("any_of", [])
            if len(req) > 0:
                here[tod.STANDARD_REQUIRED_KEY] = req
            if len(opt) > 0:
                here[tod.STANDARD_OPTIONAL_KEY] = opt
            result.append(here)
        return result

    def _get_turns_from_parsed(self, user_utt, api_calls, api_resps, sys_utt):
        assert len(api_calls) == len(api_resps)
        if len(api_calls) == 0:
            api_calls = [{}]
            api_resps = [{}]
        turns = len(api_calls)
        user_utts = [SILENCE_TOKEN] * turns
        user_utts[0] = user_utt
        sys_utts = [SILENCE_TOKEN] * turns
        sys_utts[turns - 1] = sys_utt

        result = []
        for i in range(turns):
            result.append(
                tod.TodStructuredRound(
                    sys_utt=sys_utts[i],
                    api_call_machine=api_calls[i],
                    api_resp_machine=api_resps[i],
                    user_utt=user_utts[i],
                )
            )
        return result

    def setup_episodes(self, fold):
        """
        Parses into TodStructuredEpisode.
        """
        chunks, api_schema_raw = self._load_data(fold)
        api_schemas_machine = self._parse_to_api_schema(api_schema_raw)
        episodes = []
        for _, row in chunks.iterrows():
            utterances = row["utterances"][:]
            idx = 0
            rounds = []
            goal_calls = []
            if len(utterances) > 0 and utterances[0]["speaker"] == "assistant":
                (
                    idx,
                    sys_utt,
                    api_call,
                    api_resp,
                ) = self._get_utterance_and_apis_for_speaker(
                    "assistant", utterances, idx
                )

                turns = self._get_turns_from_parsed(
                    SILENCE_TOKEN, api_call, api_resp, sys_utt
                )
                for t in turns:
                    rounds.append(t)

            while idx < len(utterances):
                (
                    idx,
                    user_utt,
                    calls_user,
                    responses_user,
                ) = self._get_utterance_and_apis_for_speaker("user", utterances, idx)
                (
                    idx,
                    sys_utt,
                    calls_system,
                    responses_system,
                ) = self._get_utterance_and_apis_for_speaker(
                    "assistant", utterances, idx
                )
                api_calls = calls_user + calls_system
                api_resps = responses_user + responses_system
                goal_calls += api_calls
                turns = self._get_turns_from_parsed(
                    user_utt, api_calls, api_resps, sys_utt
                )
                for t in turns:
                    rounds.append(t)

            episode = tod.TodStructuredEpisode(
                api_schemas_machine=api_schemas_machine,
                goal_calls_machine=goal_calls,
                rounds=rounds,
                delex=self.opt.get("delex", False),
            )
            episodes.append(episode)
        return episodes

    def get_id_task_prefix(self):
        return "Taskmaster3"

    def _label_fold(self, chunks):
        return chunks.conversation_id.apply(self._h)


class SystemTeacher(Taskmaster3Parser, tod_agents.TodSystemTeacher):
    pass


class UserSimulatorTeacher(Taskmaster3Parser, tod_agents.TodUserSimulatorTeacher):
    pass


class DefaultTeacher(SystemTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Task Oriented Dialogue (TOD) teacher metrics.
"""
from typing import Optional, List, Dict, Any
from parlai.core.metrics import AverageMetric, BleuMetric, F1Metric, Metric, Metrics


class SlotMetrics(Metrics):
    """
    Helper container which encapsulates standard slot metrics in task oriented learning
    (jga, slot_p, slot_r, etc).

    Due to differences in dialogue representations between tasks, the input is pre-
    parsed ground truth and predicted slot dictionaries.
    """

    def __init__(
        self,
        teacher_slots: Dict[str, str],
        predicted_slots: Dict[str, str],
        prefixes: Optional[List] = None,
        shared: Dict[str, Any] = None,
    ) -> None:
        super().__init__(shared=shared)
        self.prefixes = prefixes if prefixes else []
        self.add_with_prefixes("jga", AverageMetric(teacher_slots == predicted_slots))
        if len(teacher_slots) > 0:
            self.add_with_prefixes(
                "jga_noempty", AverageMetric(teacher_slots == predicted_slots)
            )
        else:
            self.add_with_prefixes(
                "jga_empty", AverageMetric(teacher_slots == predicted_slots)
            )

        # precision
        for pred_slot_name, pred_value in predicted_slots.items():
            slot_p = AverageMetric(teacher_slots.get(pred_slot_name) == pred_value)
            self.add_with_prefixes("slot_p", slot_p)
            self.add_with_prefixes("slot_f1", SlotF1Metric(slot_p=slot_p))
        # recall
        for teacher_slot_name, teacher_value in teacher_slots.items():
            slot_r = AverageMetric(
                predicted_slots.get(teacher_slot_name) == teacher_value
            )
            self.add_with_prefixes("slot_r", slot_r)
            self.add_with_prefixes("slot_f1", SlotF1Metric(slot_r=slot_r))

    def add_with_prefixes(self, name, value):
        self.add(name, value)
        for prefix in self.prefixes:
            self.add(f"{prefix}/{name}", value)


class NlgMetrics(Metrics):
    """
    Helper container for generation version of standard metrics (F1, BLEU, ..).
    """

    def __init__(
        self,
        guess: str,
        labels: Optional[List[str]],
        prefixes: Optional[List[str]] = None,
        shared: Dict[str, Any] = None,
    ) -> None:
        super().__init__(shared=shared)
        self.prefixes = prefixes if prefixes else []
        bleu = BleuMetric.compute(guess, labels)
        f1 = F1Metric.compute(guess, labels)
        self.add_with_prefixes("nlg_bleu", bleu)
        self.add_with_prefixes("nlg_f1", f1)

    def add_with_prefixes(self, name, value):
        self.add(name, value)
        for prefix in self.prefixes:
            self.add(f"{prefix}/{name}", value)


AverageType = Optional[AverageMetric]


def _average_type_sum_helper(first: AverageType, second: AverageType) -> AverageType:
    """
    Helper to deal with Nones.

    We are "clever" in how we aggregate SlotF1Metrics (See SlotMetrics `__init__`) in
    that we add precision and recall values separately, but this means we need to handle
    None.
    """
    if first is None:
        return second
    if second is None:
        return first
    return first + second


class SlotF1Metric(Metric):
    """
    Metric to keep track of slot F1.

    Keeps track of slot precision and slot recall as running metrics.
    """

    __slots__ = ("_slot_p", "_slot_r")

    @property
    def macro_average(self) -> bool:
        """
        Indicates whether this metric should be macro-averaged when globally reported.
        """
        return True

    def __init__(self, slot_p: AverageType = None, slot_r: AverageType = None):
        if not isinstance(slot_p, AverageMetric) and slot_p is not None:
            slot_p = AverageMetric(slot_p)
        if not isinstance(slot_r, AverageMetric) and slot_r is not None:
            slot_r = AverageMetric(slot_r)
        self._slot_p = slot_p
        self._slot_r = slot_r

    def __add__(self, other: Optional["SlotF1Metric"]) -> "SlotF1Metric":
        # NOTE: hinting can be cleaned up with "from __future__ import annotations" when
        # we drop Python 3.6
        if other is None:
            return self
        slot_p = _average_type_sum_helper(self._slot_p, other._slot_p)
        slot_r = _average_type_sum_helper(self._slot_r, other._slot_r)
        return type(self)(slot_p=slot_p, slot_r=slot_r)

    def value(self) -> float:
        if self._slot_p is None or self._slot_r is None:
            return float("nan")
        else:
            slot_p = self._slot_p.value()
            slot_r = self._slot_r.value()
            if slot_p == 0.0 and slot_r == 0.0:
                return float("nan")
            else:
                return 2 * (slot_p * slot_r) / (slot_p + slot_r)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Tests tod world metrics in the full script, *without* making the script properly set up
the agents on its own.

Use a few of the API Call + goal hit metrics as the metric handlers to test proper
functionality.
"""

import copy
import unittest

import parlai.core.tod.tod_test_utils.test_agents as test_agents
import parlai.scripts.tod_world_script as tod_world_script
from parlai.core.tod.tod_agents import StandaloneApiAgent
from parlai.core.tod.world_metrics_handlers import METRICS_HANDLER_CLASSES_TEST_REGISTRY
from parlai.core.metrics import dict_report

# Ignore lint on following line; want to have registered classes show up for tests
import projects.tod_simulator.world_metrics.extended_world_metrics  # noqa: F401


class TestTodWorldScript(tod_world_script.TodWorldScript):
    """
    Wrap around it to check its logic; also makes it easier to do things w/ underlying
    World.
    """

    def _get_tod_agents(self, opt):
        """
        Hack so we can separate out logic of making sure agent parsing is correct.
        """
        if hasattr(self, "agents"):
            return self.agents
        return super()._get_tod_agents(opt)

    def _setup_world(self):
        world = super()._setup_world()
        for i in range(len(world.batch_tod_world_metrics)):
            world.batch_tod_world_metrics[i].handlers = [
                x() for x in METRICS_HANDLER_CLASSES_TEST_REGISTRY
            ]
        return world

    def _save_outputs(self, opt, world, logger, episode_metrics):
        self.world = world
        self.episode_metrics = episode_metrics


class TodWorldInScriptTestBase(unittest.TestCase):
    def add_tod_world_opts(self, base_opts):
        """
        Convenience since we're initing the opt directly without parlai parser.
        """
        opts = copy.deepcopy(base_opts)
        opts["datatype"] = "DUMMY"
        opts["datafile"] = "DUMMY"
        opts["standalone_api_file"] = test_agents.API_DATABASE_FILE
        opts["exact_api_call"] = True
        opts["log_keep_fields"] = "all"
        opts["display_examples"] = False
        opts[
            "include_api_schemas"
        ] = True  # do this to test_agents.make sure they're done correctly.
        return opts

    def setup_agents(self, added_opts):
        full_opts = self.add_tod_world_opts(added_opts)
        sys = test_agents.ApiCallAndSysUttAgent(full_opts)
        agents = [
            test_agents.UserUttAgent(full_opts),
            sys,
            StandaloneApiAgent(full_opts),
            sys,
            test_agents.ApiSchemaAgent(full_opts),
            test_agents.GoalAgent(full_opts),
        ]
        return agents, full_opts

    def _run_test(self):
        self._run_test_helper(test_agents.EPISODE_SETUP__SINGLE_API_CALL)
        self._run_test_helper(test_agents.EPISODE_SETUP__MULTI_ROUND)
        self._run_test_helper(test_agents.EPISODE_SETUP__MULTI_EPISODE)
        self._run_test_helper(test_agents.EPISODE_SETUP__MULTI_EPISODE_BS)

    def _run_test_helper(self, config_base):
        config = copy.deepcopy(config_base)
        config["use_broken_mock_api_calls"] = True
        add = self.config_args()
        for key in add:
            config[key] = add[key]
        agents, opt = self.setup_agents(config)
        script = TestTodWorldScript(opt)
        script.agents = agents
        script.run()
        self._check_metrics_correct(script, opt)

    def _check_metrics_correct(self, script, opt):
        """
        Last argument is only relevant for the max_turn test.
        """
        max_rounds = opt[test_agents.TEST_NUM_ROUNDS_OPT_KEY]
        max_episodes = opt[test_agents.TEST_NUM_EPISODES_OPT_KEY]
        episode_metrics = script.episode_metrics
        for episode_idx, episode in enumerate(episode_metrics):
            goal, episode_metric = episode
            episode_metric = dict_report(episode_metric.report())
            self.assertAlmostEqual(
                episode_metric["all_goals_hit"],
                not test_agents.episode_has_broken_api_turn(episode_idx, max_rounds),
            )
        broken_episodes = sum(
            [
                test_agents.episode_has_broken_api_turn(i, max_rounds)
                for i in range(max_episodes)
            ]
        )
        report = dict_report(script.world.report())
        self.assertAlmostEqual(
            report["all_goals_hit"],
            float(max_episodes - broken_episodes) / max_episodes,
        )


class TodWorldSingleBatchTest(TodWorldInScriptTestBase):
    def config_args(self):
        config = {}
        config["batchsize"] = 1
        config["max_turns"] = 10
        return config

    def test_metricsCorrect(self):
        self._run_test()


class TodWorldNonSingleBatchTest(TodWorldInScriptTestBase):
    def config_args(self):
        config = {}
        config["batchsize"] = 4
        config["max_turns"] = 10
        return config

    def test_metricsCorrect(self):
        self._run_test()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Classifier Gpt2 Re-Ranker Object.

Provided with a classifier model file, the gpt2 re-ranker provides an API for re-ranking
candidate outputs based on maximizing the probability of a given provided class.
"""
from parlai.agents.reranker.reranker import AbstractGpt2RerankAgent
from parlai.agents.reranker.classifier_reranker import ClassifierReranker


class ClassifierGpt2Reranker(ClassifierReranker):
    pass


class ClassifierGpt2RerankerAgent(AbstractGpt2RerankAgent):
    """
    Generative GPT2 Re-rank agent for adding a ClassifierReranker.
    """

    @classmethod
    def get_reranker_class(cls):
        return ClassifierGpt2Reranker
